REQUEST_FORMATTER = "flask_logify.formatters.RequestFormatter"


def handler(formatter, **kwargs):
    return {
        "class": "logging.StreamHandler",
        "stream": "ext://sys.stderr",
        "formatter": formatter,
        **kwargs,
    }


LOGGING = dict(
    version=1,
    disable_existing_loggers=False,
    objects={"queue": {"class": "queue.Queue", "maxsize": 10000}},
    formatters={
        "simple": {"format": "[%(asctime)s][%(levelname)s]: %(message)s"},
        "consoleDebug": {
            "class": REQUEST_FORMATTER,
            "format": "[%(asctime)s]"
            "[%(levelname)s]"
            "[%(request_id)s]"
            "[%(name)s:%(module)s.%(funcName)s:%(lineno)d]: "
            "%(message)s",
        },
        "console": {
            "class": REQUEST_FORMATTER,
            "format": "[%(asctime)s][%(levelname)s][%(request_id)s]: %(message)s",
        },
        "syslog": {
            "class": REQUEST_FORMATTER,
            "format": "%(ident)s%(message)s",
        },
        "syslogNoRequest": {"format": "%(ident)s%(message)s"},
        "json": {
            "class": REQUEST_FORMATTER,
            "format": "{"
            '"requestId":"%(request_id)s",'
            '"level":"%(levelname)s",'
            '"datetime":"%(asctime)s",'
            '"message":%(message)s'
            "}",
        },
    },
    handlers={
        "simple": handler("simple"),
        "console": handler("console"),
        "consoleDebug": handler("consoleDebug"),
        "syslog": {
            "class": "flask_logify.handlers.FlaskSysLogHandler",
            "address": ["localhost", 514],
            "formatter": "syslog",
            "facility": "user",
        },
        "syslogNoRequest": {
            "class": "flask_logify.handlers.FlaskSysLogHandler",
            "address": ["localhost", 514],
            "formatter": "syslogNoRequest",
            "facility": "user",
        },
        "queueConsole": {
            "respect_handler_level": True,
            "class": "flask_logify.handlers.QueueHandler",
            "queue": "cfg://objects.queue",
            "handlers": ["cfg://handlers.console"],
        },
        "queueSyslogNoRequest": {
            "respect_handler_level": True,
            "class": "flask_logify.handlers.QueueHandler",
            "queue": "cfg://objects.queue",
            "handlers": ["cfg://handlers.syslogNoRequest"],
        },
    },
    loggers={
        "development": {
            "level": "DEBUG",
            "propagate": True,
            "handlers": ["consoleDebug"],
        },
        "developmentQueue": {
            "level": "DEBUG",
            "propagate": True,
            "handlers": ["queueConsole"],
        },
        "production": {"level": "INFO", "propagate": True, "handlers": ["console"]},
        "productionQueue": {
            "level": "INFO",
            "propagate": True,
            "handlers": ["queueConsole"],
        },
        "flask-limiter": {"level": "DEBUG", "propagate": True, "handlers": ["simple"]},
        "gunicorn.error": {
            "level": "INFO",
            "propagate": True,
            "handlers": ["queueSyslogNoRequest"],
        },
        "gunicorn.access": {
            "level": "INFO",
            "propagate": True,
            "handlers": ["queueSyslogNoRequest"],
        },
    },
)

if __name__ == "__main__":  # pragma: no cover
    import json

    print(json.dumps(LOGGING))



import os
import sys
import typing as t

from flask import Blueprint
from flask.typing import AfterRequestCallable, BeforeRequestCallable
from jinja2 import ChoiceLoader, FileSystemLoader
from vbcore.db.support import SQLASupport
from vbcore.misc import random_string
from werkzeug.middleware.lint import LintMiddleware
from werkzeug.middleware.profiler import ProfilerMiddleware
from werkzeug.routing import BaseConverter

from flaskel import config, flaskel

from .converters import CONVERTERS
from .flaskel import DumpUrls, Flaskel
from .middlewares import BaseMiddleware
from .views import BaseView

StrPathLikeType = t.Union[
    str,
    os.PathLike,
    t.Sequence[t.Union[str, os.PathLike]],
]
BlueprintType = t.Union[
    Blueprint,
    t.Tuple[Blueprint],
    t.Tuple[Blueprint, t.Dict[str, t.Any]],
]
WsgiCallable = t.Callable[[dict, t.Callable], t.Callable]
WsgiCallableType = t.TypeVar("WsgiCallableType", bound=WsgiCallable)
MiddleWareType = t.Union[
    WsgiCallableType,
    t.Tuple[WsgiCallableType],
    t.Tuple[WsgiCallableType, t.Dict[str, t.Any]],
]
BaseViewType = t.TypeVar("BaseViewType", bound=BaseView)
ViewType = t.Union[
    BaseViewType,
    t.Tuple[BaseViewType],
    t.Tuple[BaseViewType, t.Dict[str, t.Any]],
    t.Tuple[BaseViewType, Blueprint, t.Dict[str, t.Any]],
]


# pylint: disable=too-many-instance-attributes
class AppBuilder:
    conf_module = config
    app_name: str = config.FLASK_APP
    secret_file: str = config.SECRET_KEY_FILE_NAME
    flask_class: t.Type[flaskel.Flaskel] = flaskel.Flaskel
    url_converters: t.Dict[str, t.Type[BaseConverter]] = CONVERTERS

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app: t.Optional[Flaskel] = None,
        *,
        conf_module: t.Optional[t.Union[str, object]] = None,
        views: t.Tuple[ViewType, ...] = (),
        folders: t.Tuple[StrPathLikeType, ...] = (),
        blueprints: t.Tuple[BlueprintType, ...] = (),
        middlewares: t.Tuple[MiddleWareType, ...] = (),
        extensions: t.Optional[t.Dict[str, t.Any]] = None,
        after_request: t.Tuple[AfterRequestCallable, ...] = (),
        before_request: t.Tuple[BeforeRequestCallable, ...] = (),
        after_create_hook: t.Optional[t.Callable[[None], None]] = None,
        converters: t.Optional[t.Dict[str, t.Type[BaseConverter]]] = None,
        version: t.Optional[str] = None,
        **options,
    ):
        """

        :param app: an instance of Flaskel
        :param conf_module: python module file
        :param converters: custom url converter mapping
        :param extensions: custom extension appended to defaults
        :param blueprints: list of application's blueprints
        :param folders: list of custom jinja2's template folders
        :param middlewares: list of wsgi middleware
        :param views: list of views to register
        :param after_request: list of functions to register after request
        :param before_request: list of functions to register before request
        :param after_create_callback: a function called after all components are registered
        :param options: passed to Flask class
        :param version: app version, may be used by extensions
        :return:
        """
        self._app = app
        self._version = version
        self._options = options or {}
        self._converters = converters or {}
        self._extensions = extensions or {}
        self._views = views
        self._folders = folders
        self._blueprints = blueprints
        self._middlewares = middlewares
        self._after_request = after_request
        self._before_request = before_request
        self._after_create_callback = after_create_hook
        self._app_conf_module = conf_module

    @property
    def app(self):
        return self._app

    def generate_secret_key(self, secret_file, key_length):
        secret_key = random_string(key_length)

        with open(secret_file, "w", encoding="utf-8") as f:
            f.write(secret_key)
            abs_file = os.path.abspath(secret_file)
            self._app.logger.warning(
                "new secret key generated: take care of this file: %s", abs_file
            )
        return secret_key

    def load_secret_key(self, secret_file):
        if not os.path.isfile(secret_file):
            return None

        with open(secret_file, "r", encoding="utf-8") as f:
            secret_key = f.read()
            self._app.logger.info("load secret key from: %s", secret_file)
        return secret_key

    def set_secret_key(self):
        secret_key = None
        key_length = self._app.config.SECRET_KEY_MIN_LENGTH or 256

        if not self._app.config.get("FLASK_ENV", "").lower().startswith("dev"):
            secret_file = self._app.config.SECRET_KEY
            if secret_file:
                secret_key = self.load_secret_key(secret_file) or secret_file
            else:
                secret_key = self.load_secret_key(self.secret_file)
                secret_key = secret_key or self.generate_secret_key(
                    self.secret_file, key_length
                )
        elif not self._app.config.SECRET_KEY:
            self._app.logger.debug("set secret key in development mode")
            secret_key = "fake_very_complex_string"

        self._app.config.SECRET_KEY = secret_key or self._app.config.SECRET_KEY
        if len(secret_key) < key_length:
            self._app.logger.warning("secret key length is less than: %s", key_length)

        if not self._app.config.JWT_SECRET_KEY:
            self._app.config.JWT_SECRET_KEY = self._app.config.SECRET_KEY

    def set_config(self, conf: t.Optional[t.Union[str, t.Dict[str, t.Any]]] = None):
        self._app.config.from_object(self.conf_module)
        if self._app_conf_module:
            self._app.config.from_object(self._app_conf_module)
        if isinstance(conf, str):
            self._app.config.from_object(conf)
        else:
            self._app.config.from_mapping(**(conf or {}))
        self._app.config.from_envvar("APP_CONFIG_FILE", silent=True)

    @staticmethod
    def normalize_tuple(tokens: t.Union[t.Any, tuple]) -> t.Tuple[t.Any, t.Dict]:
        if not isinstance(tokens, tuple):
            return tokens, {}
        return tokens[0], tokens[1] if len(tokens) > 1 else {}

    def register_extensions(self):
        with self._app.app_context():
            for name, e in self._extensions.items():
                try:
                    ext, opt = self.normalize_tuple(e)
                    if not ext:
                        raise TypeError("extension could not be None or empty")
                except (TypeError, IndexError) as exc:
                    self._app.logger.warning(
                        "Invalid extension '%s' configuration '%s': %s", name, e, exc
                    )
                    continue

                ext.init_app(self._app, **opt)
                self._app.logger.debug(
                    "Registered extension '%s' with options: %s", name, opt
                )

        self._app.logger.debug("Dump flask extensions:")
        for k, v in self._app.extensions.items():
            self._app.logger.debug("Registered extension '%s': %s", k, v)

    def register_blueprints(self):
        for b in self._blueprints:
            bp, opt = self.normalize_tuple(b)
            self._app.register_blueprint(bp, **opt)
            self._app.logger.debug(
                "Registered blueprint '%s' with options: %s", bp.name, opt
            )

    def register_converters(self):
        self._app.url_map.converters.update(self.url_converters)
        self._app.url_map.converters.update(self._converters)
        for k, v in self._app.url_map.converters.items():
            self._app.logger.debug("Registered converter: '%s' = %s", k, v.__name__)

    def register_template_folders(self):
        loaders = [self._app.jinja_loader]

        for fsl in self._folders:
            loaders.append(FileSystemLoader(fsl))

        if self._folders:
            self._app.jinja_loader = ChoiceLoader(loaders)
            for f in self._folders:
                self._app.logger.debug("Registered template folder: '%s'", f)

    def register_middlewares(self):
        for m in self._middlewares:
            m, kwargs = self.normalize_tuple(m)
            if isinstance(m, BaseMiddleware):
                m.flask_app = self._app

            self._app.wsgi_app = m(self._app.wsgi_app, **kwargs)
            self._app.logger.debug("Registered middleware: '%s'", m.__name__)

    def register_views(self):
        def normalize_args(data: ViewType) -> tuple:
            if not isinstance(data, tuple):
                data = (data,)
            d = data + (None,) * (3 - len(data))
            if isinstance(d[1], dict):
                return d[0], d[2], d[1]
            return d[0], d[1], d[2] or {}

        def normalize_params(_v, _p: dict) -> dict:
            _p.setdefault("name", _v.default_view_name)
            _p.setdefault("urls", _v.default_urls)
            return _p

        for view in self._views:
            v, b, p = normalize_args(view)
            v.register(b or self._app, **normalize_params(v, p))
            mess = f"on blueprint '{b.name}'" if b else ""
            self._app.logger.debug(
                "Registered view: '%s' %s with params: %s", v.__name__, mess, p
            )

    def register_after_request(self):
        for f in self._after_request:
            self._app.after_request_funcs.setdefault(None, []).append(f)
            self._app.logger.debug(
                "Registered function after request: '%s'", f.__name__
            )

    def register_before_request(self):
        for f in self._before_request:
            self._app.before_request_funcs.setdefault(None, []).append(f)
            self._app.logger.debug(
                "Registered function before request: '%s'", f.__name__
            )

    def set_linter_and_profiler(self):
        if self._app.config.WSGI_WERKZEUG_PROFILER_ENABLED:
            file = self._app.config.WSGI_WERKZEUG_PROFILER_FILE
            self._app.wsgi_app = ProfilerMiddleware(
                self._app.wsgi_app,
                # pylint: disable=consider-using-with
                stream=open(file, "w", encoding="utf-8") if file else sys.stdout,
                restrictions=self._app.config.WSGI_WERKZEUG_PROFILER_RESTRICTION,
            )
            self._app.logger.debug("Registered: '%s'", ProfilerMiddleware.__name__)

        if self._app.config.WSGI_WERKZEUG_LINT_ENABLED:
            self._app.wsgi_app = LintMiddleware(self._app.wsgi_app)
            self._app.logger.debug("Registered: '%s'", LintMiddleware.__name__)

    def dump_urls(self):
        if self._app is not None:
            self._app.logger.debug("Registered routes:\n%s", DumpUrls(self._app))

    def init_db(self):
        try:
            sqlalchemy = self._app.extensions.get("sqlalchemy")
            if sqlalchemy is not None:
                SQLASupport.register_custom_handlers(sqlalchemy.db.engine)
                sqlalchemy.db.create_all()
        except Exception as exc:  # pylint: disable=broad-except
            self._app.logger.exception(exc)

    def after_create_hook(self):
        self.dump_urls()

        if self._app.debug:
            self.set_linter_and_profiler()

        with self._app.app_context():
            self.init_db()
            if callable(self._after_create_callback):
                self._after_create_callback()

    def create(
        self, conf: t.Optional[t.Union[str, t.Dict[str, t.Any]]] = None
    ) -> Flaskel:
        self._app = self._app or self.flask_class(self.app_name, **self._options)
        self._app.version = self._version

        self.set_config(conf)
        self.set_secret_key()

        self.register_extensions()
        self.register_middlewares()
        self.register_template_folders()
        self.register_converters()
        self.register_views()
        self.register_blueprints()
        self.register_after_request()
        self.register_before_request()

        self.after_create_hook()
        return self._app



import typing as t
from dataclasses import dataclass
from math import ceil

from flask import current_app as cap


class ConfigProxy:
    def __init__(self, key: t.Optional[str] = None, default=None):
        self.__key = key
        self.__default = default

    def __getattr__(self, item):
        obj = self._proxy()
        if obj == self.__default:
            return self.__default
        if not isinstance(item, str):
            return obj.get(item)
        for i in item.split("."):
            obj = obj.get(i)
        return obj

    def __call__(self, item: t.Optional[str] = None, **kwargs):
        return self.__getitem__(item)

    def __getitem__(self, item: t.Optional[str] = None):
        if item is None:
            return self._proxy()
        return self.__getattr__(item)

    def _proxy(self):
        res = cap.config

        if self.__key is None:
            return cap.config

        if "." not in self.__key:
            res = res.get(self.__key, self.__default)
        else:
            for key in self.__key.split("."):
                res = res.get(key, {})

        return res if res is not None else self.__default

    def get(self, item: t.Optional[str] = None):
        res = self(item)
        return res if res is not None else self.__default


class ExtProxy:
    def __init__(self, name: str):
        self._name = name

    @property
    def extension(self):
        attrs = self._name.split(".")
        ret = cap.extensions[attrs[0]]
        for a in attrs[1:]:
            ret = getattr(ret, a, None)

        return ret

    def __getattr__(self, item):
        return getattr(self.extension, item)

    def __call__(self, *args, **kwargs):
        return self.extension()

    def __getitem__(self, item: str):
        return self.extension[item]


@dataclass(frozen=True)
class Pagination:
    page: t.Optional[int] = None
    page_size: t.Optional[int] = None
    max_page_size: t.Optional[int] = None

    def is_paginated(self, total: int) -> bool:
        return (self.page or 1) < self.pages(total)

    def pages(self, total: int) -> int:
        page_size = self.per_page()
        return int(ceil(total / page_size)) if page_size else 0

    def offset(self) -> int:
        if not self.page:
            return 0
        return (self.page - 1) * self.per_page()

    def per_page(self) -> int:
        if self.page_size:
            if self.max_page_size:
                return min(self.page_size, self.max_page_size)
            return self.page_size
        return self.max_page_size



from flaskel.utils.schemas.default import SCHEMAS as DEFAULT_SCHEMAS

LOCALE = "it_IT.utf8"
ERROR_PAGE = "errors.html"
ERROR_HANDLER = "web"

SCHEMAS = dict(
    PostRevokeToken=DEFAULT_SCHEMAS.POST_REVOKE_TOKEN,
    PostAccessToken=DEFAULT_SCHEMAS.POST_ACCESS_TOKEN,
    AccessToken=DEFAULT_SCHEMAS.ACCESS_TOKEN,
    RefreshToken=DEFAULT_SCHEMAS.REFRESH_TOKEN,
    ApiProblem=DEFAULT_SCHEMAS.API_PROBLEM,
)

if __name__ == "__main__":  # pragma: no cover
    import json

    print(json.dumps(SCHEMAS))



import os.path
import typing as t

import flask
from vbcore.datastruct import ObjectDict
from vbcore.datastruct.lazy import Dumper
from vbcore.http import httpcode
from vbcore.http.headers import HeaderEnum
from vbcore.json import JsonDecoder, JsonEncoder
from vbcore.types import OptStr
from werkzeug.routing import Rule
from werkzeug.utils import safe_join

from flaskel.utils.datastruct import Pagination

cap: "Flaskel" = t.cast("Flaskel", flask.current_app)
request: "Request" = t.cast("Request", flask.request)


class Config(flask.Config, ObjectDict):
    def __init__(self, root_path: str, defaults: t.Optional[dict] = None):
        flask.Config.__init__(self, root_path, defaults)
        ObjectDict.__init__(self)


class DumpUrls(Dumper):
    def __init__(self, app: flask.Flask):
        super().__init__(app.url_map.iter_rules())

    @property
    def rules(self) -> t.Iterator[Rule]:
        return self.data

    def dump(self) -> str:
        output = []
        for rule in self.rules:
            methods = ",".join(sorted(rule.methods))
            output.append(f"{rule.endpoint:35s} {methods:40s} {rule}")
        return "\n".join(sorted(output))


class Request(flask.Request):
    @property
    def id(self) -> t.Optional[str]:  # pylint: disable=invalid-name
        hdr = cap.config.REQUEST_ID_HEADER
        if hasattr(flask.g, "request_id"):
            return flask.g.request_id
        if hdr in flask.request.headers:
            return flask.request.headers[hdr]

        if not hdr:
            return None

        flask_header_name = f"HTTP_{hdr.upper().replace('-', '_')}"
        return flask.request.environ.get(flask_header_name)

    def get_json(self, *args, allow_empty=False, **kwargs) -> ObjectDict:
        payload = super().get_json(*args, **kwargs)
        if payload is None:
            if not allow_empty:
                flask.abort(httpcode.BAD_REQUEST, "No JSON in request")
            payload = ObjectDict()  # pragma: no cover

        return ObjectDict.normalize(payload)


class Response(flask.Response):
    @classmethod
    def no_content(cls, status=httpcode.NO_CONTENT, headers=None) -> "Response":
        response = flask.make_response(bytes())
        response.headers.update(headers or {})
        response.headers.pop(HeaderEnum.CONTENT_TYPE)
        response.headers.pop(HeaderEnum.CONTENT_LENGTH)
        response.status_code = status
        return t.cast("Response", response)

    @classmethod
    def set_sendfile_headers(cls, response: "Response", file_path: str) -> "Response":
        hdr = HeaderEnum
        conf = cap.config
        response.headers[hdr.X_ACCEL_REDIRECT] = os.path.abspath(file_path)
        response.headers[hdr.X_ACCEL_CHARSET] = conf.ACCEL_CHARSET or "utf-8"
        response.headers[hdr.X_ACCEL_BUFFERING] = (
            "yes" if conf.ACCEL_BUFFERING else "no"
        )
        if conf.ACCEL_LIMIT_RATE:
            response.headers[hdr.X_ACCEL_LIMIT_RATE] = conf.ACCEL_LIMIT_RATE
        if conf.SEND_FILE_MAX_AGE_DEFAULT:
            response.headers[hdr.X_ACCEL_EXPIRES] = conf.SEND_FILE_MAX_AGE_DEFAULT
        return response

    @classmethod
    def send_file(cls, directory: str, filename: str, **kwargs) -> "Response":
        kwargs.setdefault("as_attachment", True)
        file_path = safe_join(directory, filename)

        try:
            response = flask.send_file(file_path, etag=True, conditional=True, **kwargs)
        except IOError as exc:
            cap.logger.warning(str(exc))
            return flask.abort(httpcode.NOT_FOUND)

        if cap.config.USE_X_SENDFILE is True and cap.config.ENABLE_ACCEL is True:
            # following headers works with nginx compatible proxy
            return cls.set_sendfile_headers(response, file_path)  # type: ignore
        return response  # type: ignore

    def get_json(self, *args, **kwargs) -> ObjectDict:
        return ObjectDict.normalize(super().get_json(*args, **kwargs))

    @classmethod
    def pagination_headers(cls, total: int, pagination: Pagination) -> t.Dict[str, int]:
        return {
            HeaderEnum.X_PAGINATION_COUNT: total,
            HeaderEnum.X_PAGINATION_PAGE: pagination.page or 1,
            HeaderEnum.X_PAGINATION_NUM_PAGES: pagination.pages(total),
            HeaderEnum.X_PAGINATION_PAGE_SIZE: pagination.per_page(),
        }


class Flaskel(flask.Flask):
    config_class = Config
    request_class = Request
    response_class = Response
    _json_encoder = JsonEncoder
    _json_decoder = JsonDecoder

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.version: OptStr = None
        self.config: Config



from redis import StrictRedis
from vbcore.date_helper import Seconds


class FlaskRedis:
    def __init__(self, app=None, client=None, **kwargs):
        self._client = client

        if app is not None:
            self.init_app(app, **kwargs)

    @property
    def client(self):
        return self._client

    def init_app(self, app, redis_class=StrictRedis, **kwargs):
        app.config.setdefault("REDIS_URL", "redis://localhost:6379/0")
        app.config.setdefault("REDIS_OPTS", {})
        app.config["REDIS_OPTS"].setdefault("decode_responses", True)
        app.config["REDIS_OPTS"].setdefault("socket_connect_timeout", Seconds.millis)
        app.config["REDIS_OPTS"].update(**kwargs)

        if not self._client:
            self._client = redis_class.from_url(
                app.config["REDIS_URL"], **app.config["REDIS_OPTS"]
            )

        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["redis"] = self

    def __getattr__(self, name):
        return getattr(self.client, name)



try:
    from gevent import monkey

    monkey.patch_all()
except ImportError:
    try:
        from eventlet import monkey_patch

        monkey_patch()
    except ImportError:
        pass

from typing import Dict

import click

from flaskel import AppBuilder, middlewares as middle
from flaskel.ext.websocket import socketio

EXTENSIONS: Dict[str, Dict] = {}

APP_CONFIG = dict(
    static_folder=None,
    extensions=EXTENSIONS,
    middlewares=(
        middle.RequestID,
        middle.ReverseProxied,
    ),
)


def get_app():
    return AppBuilder(**APP_CONFIG).get_or_create()


@click.command()
@click.option("-p", "--port", default=5000, type=int)
@click.option("-H", "--bind", default="127.0.0.1")
def cli(port, bind):
    app = get_app()
    socketio.run(app, host=bind, port=port, debug=app.debug, log_output=app.debug)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



import typing as t

from requests import exceptions as http_exc
from vbcore.http.batch import HTTPBatch
from vbcore.http.client import HTTPClient, JsonRPCClient
from vbcore.http.httpdumper import LazyHTTPDumper
from vbcore.uuid import get_uuid

from flaskel.flaskel import cap, request

HTTPStatusError = (http_exc.HTTPError,)
NetworkError = (http_exc.ConnectionError, http_exc.Timeout)
all_errors = (*HTTPStatusError, *NetworkError)


class FlaskelHTTPDumper(LazyHTTPDumper):
    @classmethod
    def dump_request(cls, req, *_, dump_body=None, **kwargs):
        h = cap.config.LOG_REQ_HEADERS
        return super().dump_request(req, dump_body, only_hdr=h, **kwargs)

    @classmethod
    def dump_response(cls, resp, *_, dump_body=None, **kwargs):
        h = cap.config.LOG_RESP_HEADERS
        return super().dump_response(resp, dump_body, only_hdr=h, **kwargs)


class FlaskelHttpBatch(FlaskelHTTPDumper, HTTPBatch):
    def __init__(self, endpoint: t.Optional[str] = None, **kwargs):
        kwargs.setdefault("logger", cap.logger)
        kwargs.setdefault("timeout", cap.config.HTTP_TIMEOUT or 10)
        super().__init__(endpoint, **kwargs)

    def request(self, requests, **kwargs):
        if request.id:
            for r in requests:
                if not r.get("headers"):
                    r["headers"] = {}
                req_id = f"{request.id},{get_uuid()}"
                r["headers"][cap.config.REQUEST_ID_HEADER] = req_id

        kwargs.setdefault("verify", cap.config.HTTP_SSL_VERIFY)
        return super().request(requests, **kwargs)


class FlaskelHttp(FlaskelHTTPDumper, HTTPClient):
    def __init__(self, endpoint, *args, **kwargs):
        kwargs.setdefault("logger", cap.logger)
        super().__init__(endpoint, *args, **kwargs)
        self._timeout = cap.config.HTTP_TIMEOUT or self._timeout

    def request(self, uri, **kwargs):
        if request.id:
            if not kwargs.get("headers"):
                kwargs["headers"] = {}
            kwargs["headers"][cap.config.REQUEST_ID_HEADER] = request.id

        kwargs.setdefault("verify", cap.config.HTTP_SSL_VERIFY)
        return super().request(uri, **kwargs)


class FlaskelJsonRPC(FlaskelHttp, JsonRPCClient):
    pass



from flask import current_app as cap
from flask_errors_handler import DefaultNormalizer
from flask_jwt_extended.exceptions import JWTExtendedException
from jwt import PyJWTError

from flaskel.http.exceptions import Unauthorized

jwt_errors = (JWTExtendedException, PyJWTError)


class ErrorNormalizer(DefaultNormalizer):
    def normalize(self, ex, *_, **__):
        if isinstance(ex, jwt_errors):
            response = {}
            if cap.debug:
                response["message"] = str(ex)
                response["exception"] = ex.__class__.__name__
            ex = Unauthorized(response=response)

        return super().normalize(ex)



import typing as t

from flask import stream_with_context
from vbcore import uuid
from vbcore.datastruct import ObjectDict
from vbcore.http import httpcode, HttpMethod
from vbcore.http.client import HTTPBase
from vbcore.http.headers import ContentTypeEnum, HeaderEnum
from vbcore.http.rpc import rpc_error_to_httpcode

from flaskel import abort, cap, flaskel, request
from flaskel.http.client import FlaskelHttp, FlaskelJsonRPC

from .base import BaseView


class ProxyView(BaseView):
    client_class: t.Type[HTTPBase] = FlaskelHttp

    def __init__(
        self,
        host: t.Optional[str] = None,
        url: t.Optional[str] = None,
        method: t.Optional[str] = None,
        proxy_body: bool = False,
        proxy_headers: bool = False,
        proxy_params: bool = False,
        stream: bool = True,
        skip_args: t.Tuple[str, ...] = (),
        options: t.Optional[t.Callable] = None,
        **kwargs,
    ):
        """

        :param host:
        :param url:
        :param method:
        :param proxy_body:
        :param proxy_headers:
        :param proxy_params:
        :param stream: if False streaming response are disabled
        :param skip_args: a tuple o arguments name to remove from kwargs
                         it is necessary because flask pass url params to dispatch_request
        :param options: callable that returns dict to pass to http client instance
                        It overrides kwargs
        """
        self._host = host
        self._method = method
        self._url = url
        self._proxy_body = proxy_body
        self._proxy_headers = proxy_headers
        self._proxy_params = proxy_params
        self._options = kwargs
        self._skip_args = skip_args
        self._stream = stream

        if callable(options):
            self._options = {**kwargs, **options()}  # pragma: no cover

    def _filter_kwargs(self, data: dict) -> dict:
        for arg in self._skip_args:
            data.pop(arg, None)
        return data

    def dispatch_request(self, *_, **kwargs):
        opts = {**self._options, **self._filter_kwargs(kwargs)}
        response = self.proxy(self.service(), **opts)

        if response and response.body and response.status != httpcode.NO_CONTENT:
            if self._stream:
                return flaskel.Response(
                    stream_with_context(response.body),
                    status=response.status,
                    headers=response.headers,
                )
            return response.body, response.status, response.headers
        return flaskel.Response.no_content()

    def proxy(self, data: ObjectDict, **kwargs) -> ObjectDict:
        client = self.client_class(data.host or self.upstream_host(), **kwargs)
        return client.request(
            data.url or self.request_url(),
            method=data.method or self.request_method(),
            headers=data.headers or self.request_headers(),
            params=data.params or self.request_params(),
            data=data.body or self.request_body(),
            stream=self._stream,
        )

    def service(self) -> ObjectDict:
        return ObjectDict(
            host=self.upstream_host(),
            uri=self.request_url(),
            method=self.request_method(),
            headers=self.request_headers(),
            params=self.request_params(),
            data=self.request_body(),
        )

    def upstream_host(self) -> str:
        return self._host

    def request_url(self) -> str:
        return self._url or request.path

    def request_method(self) -> str:
        return self._method or request.method

    def request_body(self):
        return request.get_data() if self._proxy_body else None

    def request_headers(self) -> t.Optional[t.Dict[str, t.Any]]:
        return dict(request.headers.items()) if self._proxy_headers else None

    def request_params(self) -> t.Optional[t.Dict[str, t.Any]]:
        return request.args if self._proxy_params else None


class ConfProxyView(BaseView):
    sep = "."
    config_key = None

    def __init__(self, config_key: t.Optional[str] = None):
        self._config_key = config_key or self.config_key
        assert self._config_key is not None, "missing 'config_key'"

    def dispatch_request(self, *args, **kwargs):
        return self.perform(*args, **kwargs)

    def perform(self, *_, item=None, **__) -> ObjectDict:
        conf = cap.config
        keys = self._config_key.split(self.sep)
        if item is not None:
            keys += item.split(self.sep)
        for k in keys:
            conf = conf.get(k) or ObjectDict()
        if not conf:
            cap.logger.warning(f"unable to find conf keys: {self.sep.join(keys)}")
            abort(httpcode.NOT_FOUND)
        return conf


class TransparentProxyView(ProxyView):
    methods: t.ClassVar[t.Optional[t.Collection[str]]] = [
        HttpMethod.POST,
        HttpMethod.PUT,
        HttpMethod.GET,
        HttpMethod.DELETE,
    ]

    def __init__(
        self,
        host: t.Optional[str] = None,
        url: t.Optional[str] = None,
        method: t.Optional[str] = None,
        proxy_body: bool = True,
        proxy_headers: bool = True,
        proxy_params: bool = True,
        stream: bool = True,
        skip_args: t.Tuple[str, ...] = (),
        options: t.Optional[t.Callable] = None,
        **kwargs,
    ):
        super().__init__(
            host=host,
            url=url,
            method=method,
            proxy_body=proxy_body,
            proxy_headers=proxy_headers,
            proxy_params=proxy_params,
            skip_args=skip_args,
            stream=stream,
            options=options,
            **kwargs,
        )


class SchemaProxyView(ConfProxyView):
    default_view_name = "schema_proxy"
    default_urls = ("/schema/<path:filepath>",)

    def __init__(
        self,
        config_key: str = "SCHEMAS",
        case_sensitive: bool = False,
        ext_support: bool = True,
        ext_name: str = ".json",
    ):
        super().__init__(config_key)
        self.ext_name = ext_name
        self.ext_support = ext_support
        self.case_sensitive = case_sensitive

    def normalize(self, filepath: str) -> str:
        if self.ext_support and filepath.endswith(self.ext_name):
            filepath = self.sep.join(filepath.split(self.sep)[:-1])
        if self.case_sensitive is False:
            filepath = filepath.upper()

        return filepath.replace("/", self.sep)

    def dispatch_request(self, filepath, *args, **kwargs):
        schema_path = self.normalize(filepath)
        return super().perform(item=schema_path)


class JsonRPCProxy(ProxyView):
    default_view_name: str = "jsonrpc_proxy"
    response_content_type: str = ContentTypeEnum.JSON
    client_class: t.Type[FlaskelJsonRPC] = FlaskelJsonRPC

    methods: t.ClassVar[t.Optional[t.Collection[str]]] = [
        HttpMethod.GET,
        HttpMethod.POST,
    ]

    def __init__(
        self,
        url: t.Optional[str] = None,
        proxy_headers: bool = False,
        stream: bool = False,
        skip_args: t.Tuple[str, ...] = (),
        options: t.Optional[t.Callable] = None,
        namespace: t.Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            url=url,
            stream=stream,
            proxy_headers=proxy_headers,
            skip_args=skip_args,
            options=options,
            **kwargs,
        )
        self.namespace = namespace

    def proxy(self, data: ObjectDict, **kwargs) -> ObjectDict:
        url = self.request_url()
        client = self.client_class(url, url, **kwargs)

        if request.method == HttpMethod.GET:
            resp = client.request(
                self.request_method(),
                request_id=request.id or uuid.get_uuid(),
                stream=self._stream,
                params=self.request_params(),
                headers=self.request_headers(),
            )
            return self.prepare_response(resp)

        client.notification(
            self.request_method(),
            stream=self._stream,
            params=self.request_params(),
            headers=self.request_headers(),
        )
        return self.prepare_response()

    def prepare_response(
        self, resp: t.Optional[ObjectDict] = None, **kwargs
    ) -> ObjectDict:
        headers = {HeaderEnum.CONTENT_TYPE: self.response_content_type, **kwargs}
        status = httpcode.NO_CONTENT if resp is None else httpcode.SUCCESS

        if resp and resp.error is not None:
            status = rpc_error_to_httpcode(resp.error.code)
            abort(status, response=resp.error)

        return ObjectDict(body=resp, status=status, headers=headers)

    def request_url(self) -> str:
        return self._url

    def request_method(self) -> str:
        action = request.path.split("/")[-1]
        if self.namespace is None:
            return action
        return f"{self.namespace}.{action}"

    def request_params(self) -> dict:
        if request.method == HttpMethod.GET:
            return request.args.to_dict()
        return request.json



import typing as t

from flaskel.flaskel import Flaskel


class BaseApplication:
    default_host = "127.0.0.1"
    default_port = 5000

    def __init__(self, app: Flaskel, options: t.Optional[dict] = None):
        self.application = app
        self.options = options or {}

        default_bind = f"{self.default_host}:{self.default_port}"
        bind = (self.options.get("bind") or default_bind).split(":")

        self._interface = bind[0] or self.default_host
        self._port = int(bind[1]) if len(bind) > 1 else self.default_port
        self._bind = (self._interface, self._port)

    def run(self):
        raise NotImplementedError  # pragma: no cover


class WSGIBuiltin(BaseApplication):
    def run(self):
        debug = self.application.config.DEBUG or False
        self.application.run(host=self._interface, port=self._port, debug=debug)



from vbcore.http import httpcode

from flaskel import cap, Response
from flaskel.config import config

from .base import BaseView


class StaticFileView(BaseView):
    default_view_name = "static"
    default_static_path = "static"
    default_urls = ("/static/<path:filename>",)

    def dispatch_request(self, filename, *_, **__):
        return Response.send_file(
            filename=filename,
            max_age=cap.get_send_file_max_age,
            directory=self.default_static_path,
        )


# TODO: check me
class SPAView(BaseView):
    template = "index.html"
    default_view_name = "webapp"
    default_urls = ("/", "/<path:path>")
    default_static_folder = config("SPA_STATIC_FOLDER", default="webapp")
    default_static_url_path = config("SPA_STATIC_URL_PATH", default="/webapp")
    default_template_folder = config("SPA_TEMPLATE_FOLDER", default="templates")

    def dispatch_request(self, *_, path=None, **__):
        if path is not None:
            return Response.no_content(httpcode.NOT_MODIFIED)

        return Response.send_file(
            filename=self.template,
            max_age=cap.get_send_file_max_age,
            directory=self.default_template_folder,
        )



from .builder import AppBuilder
from .flaskel import cap, Flaskel, Request, request, Response
from .http.client import (
    FlaskelHttp,
    FlaskelHttpBatch,
    FlaskelHTTPDumper,
    FlaskelJsonRPC,
)
from .http.exceptions import abort
from .standalone import Server
from .tester import TestClient
from .utils import webargs
from .utils.datastruct import ConfigProxy, ExtProxy
from .utils.validator import PayloadValidator
from .version import *  # noqa: F403
from .wsgi import BaseApplication, WSGIFactory

client_redis = ExtProxy("redis")
client_mail = ExtProxy("client_mail")
job_scheduler = ExtProxy("scheduler")
client_mongo = ExtProxy("mongo.default.db")
db_session = ExtProxy("sqlalchemy.db.session")



import typing as t

import flask_jwt_extended
from vbcore.datastruct import ObjectDict
from vbcore.http import httpcode, HttpMethod

from flaskel import abort, cap, request, webargs
from flaskel.ext.auth import BaseTokenHandler
from flaskel.ext.default import builder
from flaskel.views import BaseView
from flaskel.views.base import UrlRule


class BaseTokenAuth(BaseView):
    jwt = flask_jwt_extended
    handler: BaseTokenHandler = BaseTokenHandler()

    default_view_name = "token_auth"

    methods: t.ClassVar[t.Optional[t.Collection[str]]] = [
        HttpMethod.GET,
        HttpMethod.POST,
    ]

    """
        NOTE: endpoint should not be change otherwise dispatch_request MUST change
    """
    default_urls = (
        UrlRule(url="/token/access", endpoint="token_access"),
        UrlRule(url="/token/refresh", endpoint="token_refresh"),
        UrlRule(url="/token/revoke", endpoint="token_revoke"),
        UrlRule(url="/token/check", endpoint="token_check"),
    )

    @classmethod
    def check_credential(cls):
        data = request.json
        if (
            data.email == cap.config.ADMIN_EMAIL
            and data.password == cap.config.ADMIN_PASSWORD
        ):
            return data
        return None

    @classmethod
    @webargs.query(
        dict(
            expires_access=webargs.OptField.positive(),
            expires_refresh=webargs.OptField.positive(),
        )
    )
    def access_token(cls, args) -> ObjectDict:
        credentials = cls.check_credential()
        if credentials:
            return cls.handler.create(credentials, **args).to_dict()
        return abort(httpcode.UNAUTHORIZED)

    @classmethod
    def refresh_token(cls) -> ObjectDict:
        @cls.jwt.jwt_required(refresh=True)
        def _refresh_token():
            return cls.handler.refresh().to_dict()

        return _refresh_token()

    @classmethod
    def check_token(cls):
        @cls.jwt.jwt_required()
        def _check_token():
            return cls.handler.dump().to_dict()

        return _check_token()

    @classmethod
    @builder.no_content
    def revoke_token(cls):
        data = request.json
        if data.access_token:
            cls.handler.revoke(data.access_token)
        if data.refresh_token:
            cls.handler.revoke(data.refresh_token)

    def dispatch_request(self, *_, **__):
        if request.method == HttpMethod.GET:
            return self.check_token()

        view_name = request.endpoint
        if view_name.endswith("token_access"):
            return self.access_token()  # pylint: disable=no-value-for-parameter
        if view_name.endswith("token_refresh"):
            return self.refresh_token()
        if view_name.endswith("token_revoke"):
            return self.revoke_token()

        return abort(httpcode.BAD_REQUEST)



from vbcore.datastruct import ObjectDict
from vbcore.jsonschema.support import Fields

from .openapi3 import SCHEMA as OPENAPI_SCHEMA

SCHEMAS = ObjectDict(
    OPENAPI=OPENAPI_SCHEMA,
    JSONRPC=ObjectDict(
        REQUEST=Fields.oneof(
            Fields.ref("/definitions/request", description="An individual request"),
            Fields.array(
                items=Fields.ref("/definitions/request"),
                description="An array of requests",
            ),
            **Fields.schema,
            description="A JSON RPC 2.0 request",
            definitions={
                "request": Fields.object(
                    required=["jsonrpc", "method"],
                    properties={
                        "jsonrpc": Fields.enum("2.0"),
                        "method": Fields.string,
                        "params": Fields.type("array", "object"),
                        "id": Fields.type(
                            "string",
                            "number",
                            "null",
                            note=[
                                "While allowed, null should be avoided: "
                                "http://www.jsonrpc.org/specification#id1",
                                "While allowed, a number with a fractional part should be avoided: "
                                "http://www.jsonrpc.org/specification#id2",
                            ],
                        ),
                    },
                )
            },
        ),
        RESPONSE=Fields.oneof(
            Fields.ref("/definitions/response"),
            Fields.array(items=Fields.ref("/definitions/response")),
            **Fields.schema,
            definitions={
                "response": Fields.type(
                    "array",
                    "object",
                    required=["jsonrpc"],
                    properties={
                        "jsonrpc": Fields.enum("2.0"),
                        "id": Fields.type("string", "number", "null"),
                        "result": Fields.type("array", "object", "null"),
                        "error": Fields.type(
                            "array",
                            "object",
                            properties={
                                "code": Fields.number,
                                "message": Fields.string,
                            },
                        ),
                    },
                )
            },
        ),
    ),
    POST_REVOKE_TOKEN=Fields.object(
        all_required=False,
        properties={
            "access_token": Fields.string,
            "refresh_token": Fields.string,
            "device_token": Fields.string,
        },
    ),
    POST_ACCESS_TOKEN=Fields.object(
        properties={"email": Fields.string, "password": Fields.string}
    ),
    ACCESS_TOKEN=Fields.object(
        required=["access_token", "refresh_token", "expires_in", "issued_at"],
        properties={
            "access_token": Fields.string,
            "refresh_token": Fields.string,
            "expires_in": Fields.Opt.integer,
            "issued_at": Fields.integer,
            "token_type": Fields.string,
            "scope": Fields.Opt.string,
        },
    ),
    REFRESH_TOKEN=Fields.object(
        required=["access_token", "expires_in", "issued_at"],
        properties={
            "access_token": Fields.string,
            "expires_in": Fields.Opt.integer,
            "issued_at": Fields.integer,
            "token_type": Fields.string,
            "scope": Fields.Opt.string,
        },
    ),
    REGISTER_CONFIRM=Fields.object(properties={"token": Fields.string}),
    PASSWORD_RESET=Fields.object(
        properties={
            "email": Fields.string,
            "new_password": Fields.string,
            "old_password": Fields.string,
        }
    ),
    PASSWORD_FORGOT=Fields.object(properties={"email": Fields.string}),
    PASSWORD_CONFIRM=Fields.object(
        properties={"token": Fields.string, "password": Fields.string}
    ),
    API_PROBLEM=Fields.object(
        properties={
            "type": Fields.string,
            "title": Fields.string,
            "detail": Fields.string,
            "instance": Fields.string,
            "status": Fields.integer,
            "response": Fields.any,
        }
    ),
    HEALTH_CHECK=Fields.object(
        **Fields.schema,
        properties={
            "status": Fields.string,
            "checks": Fields.object(
                patternProperties={
                    ".": Fields.object(
                        properties={
                            "status": Fields.string,
                            "output": Fields.type("null", "string", "object"),
                        }
                    )
                }
            ),
            "links": Fields.object(properties={"about": Fields.Opt.string}),
        },
    ),
)

if __name__ == "__main__":  # pragma: no cover
    import json

    print(json.dumps(SCHEMAS))



import click

from flaskel.scripts.init_app import init_app

cli = click.Group()


@cli.command(name="init")
@click.argument("name")
def init(name):
    """Create skeleton for new application"""
    init_app(name)


if __name__ == "__main__":
    cli()



import typing as t

from vbcore.http import httpcode, HttpMethod

from flaskel import abort, cap, request
from flaskel.views import BaseView
from flaskel.views.static import StaticFileView

from .exceptions import BadMediaError
from .service import MediaService


class ApiMedia(BaseView):
    service = MediaService
    methods: t.ClassVar[t.Optional[t.Collection[str]]] = [
        HttpMethod.POST,
        HttpMethod.DELETE,
    ]
    default_urls = (
        "/resources/<int:eid>/media",
        "/resources/<int:eid>/media/<int:res_id>",
    )

    def dispatch_request(self, eid, *_, res_id=None, **__):
        if request.method == HttpMethod.DELETE:
            self.service.delete(eid, res_id)
            return httpcode.NO_CONTENT

        try:
            files = request.files.values()
            return self.service.upload(files, eid)
        except BadMediaError as exc:
            cap.logger.exception(exc)
            return abort(httpcode.BAD_REQUEST, str(exc))


class GetMedia(StaticFileView):
    default_static_path = "static"
    default_view_name = "serve_media"
    default_urls = ("/static/media/<path:filename>",)



import typing as t

import sqlalchemy as sa
from pyfcm import FCMNotification as Firebase
from pyfcm.errors import FCMError
from sqlalchemy.exc import SQLAlchemyError
from vbcore.datastruct import ObjectDict
from vbcore.db.mixins import ExtraMixin
from vbcore.db.support import SQLASupport
from vbcore.http import HttpMethod
from vbcore.jsonschema.support import Fields
from vbcore.uuid import get_uuid

from flaskel import (
    db_session,
    ExtProxy,
    job_scheduler,
    PayloadValidator,
    request,
    Response,
)
from flaskel.views import BaseView


class SCHEMAS:
    DEVICE_REGISTER = Fields.object(properties={"token": Fields.string})
    SEND_PUSH = Fields.object(
        required=["title", "message"],
        properties={
            "title": Fields.string,
            "message": Fields.string,
            "tokens": Fields.oneof(Fields.array(items=Fields.string)),
            "user_ids": Fields.oneof(Fields.array(items=Fields.string)),
        },
    )


class DeviceModelMixin(ExtraMixin):
    token = sa.Column(sa.String(255), unique=True, nullable=False)
    user_id = sa.Column(sa.String(255), index=True, default=get_uuid())
    user_agent = sa.Column(sa.String(255), nullable=False)


class FCMNotification:
    def __init__(self, app=None, model=None, session=None, dry_run: bool = False):
        self.app = app
        self.model = model
        self.dry_run = dry_run
        self.service: Firebase
        self.sa_support: SQLASupport
        self.session = session or db_session

        if model is not None:
            self.sa_support = SQLASupport(self.model, self.session)

        if app is not None:
            self.init_app(app, model, session, dry_run)

    def init_app(self, app, model, session=None, dry_run: bool = False):
        self.app = app
        self.model = model
        self.session = session
        self.dry_run = dry_run
        self.sa_support = SQLASupport(self.model, self.session)
        self.service = Firebase(api_key=self.app.config.FCM_API_KEY)
        self.service.FCM_MAX_RECIPIENTS = self.app.config.FCM_MAX_RECIPIENTS
        app.extensions["fcm_notification"] = self

    def register_device(self, data: ObjectDict):
        try:
            self.sa_support.update_or_create(data, token=data.token)
            self.session.commit()
        except SQLAlchemyError as exc:
            self.app.logger.exception(exc)
            self.session.rollback()

    def get_tokens(self, user_ids: t.List[str]) -> t.List[str]:
        # TODO: find a way to avoid big user_ids list
        query = self.model.query.with_entities(self.model.token).where(
            self.model.user_id.in_(user_ids)
        )
        return [token[0] for token in query]

    def send_push_notification(
        self,
        title: str,
        message: str,
        tokens: t.Optional[t.List[str]] = None,
        user_ids: t.Optional[t.List[str]] = None,
        **kwargs,
    ):
        """

        :param title: notification title
        :param message: notification message
        :param user_ids: list of user ids (optional)
        :param tokens: list of fcm tokens (optional)
        :param kwargs: passed to notify_multiple_devices
        :return:
        """
        if tokens is None:
            if not user_ids:
                raise ValueError("one of 'tokens' or 'user_ids' must be given")
            try:
                tokens = self.get_tokens(user_ids)
            except SQLAlchemyError as exc:  # pragma: no cover
                self.app.logger.exception(exc)
                return None

        kwargs.setdefault("sound", "Default")
        kwargs.setdefault("dry_run", self.dry_run)

        try:
            return self.service.notify_multiple_devices(
                tokens,
                message_title=title,
                message_body=message,
                **kwargs,
            )
        except FCMError as exc:  # pragma: no cover
            self.app.logger.exception(exc)
            return None


class DeviceRegisterView(BaseView):
    methods: t.ClassVar[t.Optional[t.Collection[str]]] = [HttpMethod.POST]
    default_view_name = "fcm_device_register"
    default_urls = ("/device/register",)
    handler = ExtProxy("fcm_notification")
    schema = "DEVICE_REGISTER"

    def get_user_id(self):
        """must be implemented in subclass"""
        return self.not_implemented()

    def dispatch_request(self, *_, **__):
        payload = PayloadValidator.validate(self.schema)
        payload.user_id = self.get_user_id()
        payload.useragent = request.user_agent.string
        job_scheduler.add(
            self.handler.register_device,
            args=(payload,),
        )
        return Response.no_content()


class SendPushView(BaseView):
    methods: t.ClassVar[t.Optional[t.Collection[str]]] = [HttpMethod.POST]
    default_view_name = "fcm_send"
    default_urls = ("/send/push",)
    handler = ExtProxy("fcm_notification")
    schema = "SEND_PUSH"

    def dispatch_request(self, *_, **__):
        payload = PayloadValidator.validate(self.schema)
        job_scheduler.add(
            self.handler.send_push_notification,
            kwargs=payload,
        )
        return Response.no_content()



from flask.testing import FlaskClient

from flaskel.builder import AppBuilder


class JSONRPCMixin:
    @staticmethod
    def prepare(method=None, call_id=None, params=None, version="2.0"):
        payload = {"jsonrpc": version, "method": method, "params": params}
        if call_id:
            payload["id"] = call_id
        return payload

    def jsonrpc(
        self, url, method=None, call_id=None, params=None, version="2.0", **kwargs
    ):
        """

        :param url:
        :param method:
        :param call_id:
        :param params:
        :param version:
        :param kwargs:
        :return:
        """
        kwargs.setdefault("json", self.prepare(method, call_id, params, version))
        return self.post(url, **kwargs)

    def jsonrpc_batch(self, url, requests, version="2.0", **kwargs):
        payload = [self.prepare(**r, version=version) for r in requests]
        return self.jsonrpc(url, json=payload, **kwargs)


class TestClient(FlaskClient, JSONRPCMixin):
    def fetch(self, url, *args, **kwargs):  # pragma: no cover
        return self.open(url, method="FETCH", *args, **kwargs)

    @classmethod
    def get_app(cls, conf, **kwargs):
        """

        :param conf:
        :param kwargs:
        :return:
        """
        app = AppBuilder(**kwargs).create(conf)
        app.test_client_class = cls
        app.TESTING = True

        app.app_context().push()
        return app



import string
import typing as t

import sqlalchemy as sa
from flask import render_template
from sqlalchemy.exc import SQLAlchemyError
from vbcore.datastruct import ObjectDict
from vbcore.db.exceptions import DBDuplicateEntry
from vbcore.db.mixins import StandardMixin, UserMixin
from vbcore.http import httpcode, HttpMethod
from vbcore.jsonschema.support import Fields
from vbcore.misc import random_string

from flaskel import (
    abort,
    cap,
    client_mail,
    client_redis,
    db_session,
    ExtProxy,
    job_scheduler,
    PayloadValidator,
    request,
    Response,
)
from flaskel.views import BaseView


class AccountModel(UserMixin, StandardMixin):
    is_active = sa.Column(sa.Boolean, default=False)
    is_confirmed = sa.Column(sa.Boolean, default=False)

    def to_dict(self):
        return ObjectDict(
            email=self.email,
            is_active=self.is_active,
            created_at=self.created_at,
        )


class AuthCode:
    def __init__(self, cache, prefix: str = "auth_code//", expires: int = 600):
        self.cache = cache
        self.expires = expires
        self.key_format = f"{prefix}/{{key}}/{{token}}"

    @classmethod
    def token_value(cls, length: int = 8) -> str:
        return random_string(length, string.digits)

    def generate(self, key: str, user: AccountModel) -> str:
        token = self.token_value()
        cache_key = self.key_format.format(key=key, token=token)
        self.cache.set(cache_key, user.email, ex=self.expires)
        return token

    def check(self, key: str, token: str) -> str:
        cache_key = self.key_format.format(key=key, token=token)
        email = self.cache.get(cache_key)
        email = email.decode() if isinstance(email, bytes) else email
        self.cache.delete(cache_key)
        if not email:
            abort(httpcode.NOT_FOUND)
        return email


class AccountHandler:
    session = db_session
    scheduler = job_scheduler
    client_mail = client_mail
    auth_code_handler_class = AuthCode

    CACHE_TEMP_KEYS = ObjectDict(
        account_forgot="account_forgot",
        account_register="account_register",
    )

    NOTIFICATIONS = ObjectDict(
        register={
            "subject": "Confirm your registration",
            "body": "emails/registration.html",
        },
        forgot={
            "subject": "Forgot Password",
            "body": "emails/forgot.html",
        },
    )

    def __init__(
        self,
        app=None,
        model=AccountModel,
        auth_code_handler: t.Optional[AuthCode] = None,
        **kwargs,
    ):
        self.model = model
        self.auth_code_handler = auth_code_handler or self.auth_code_handler_class(
            client_redis, **kwargs
        )

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        app.config.setdefault("NOTIFICATIONS", self.NOTIFICATIONS)
        app.extensions["account"] = self

    @classmethod
    def sendmail(cls, recipients: t.List[str], subject: str, template: str, **kwargs):
        cls.client_mail.sendmail(
            subject=subject,
            recipients=recipients,
            html=render_template(template, **kwargs),
        )

    def prepare_user(self, data: dict):
        return self.model(**data)

    def find_by_email(self, email: str):
        # noinspection PyUnresolvedReferences
        return self.model.query.filter_by(email=email).first_or_404()

    def register(self, payload: ObjectDict):
        user = self.prepare_user(payload)

        try:
            self.session.add(user)
            self.session.commit()
        except Exception as exc:  # pylint: disable=broad-except
            cap.logger.exception(exc)
            self.session.rollback()
            if isinstance(exc, DBDuplicateEntry):
                abort(httpcode.CONFLICT, response={"cause": exc.as_dict()})
            abort(httpcode.INTERNAL_SERVER_ERROR)

        token = self.auth_code_handler.generate(
            self.CACHE_TEMP_KEYS.account_register, user
        )
        self.sendmail(
            recipients=[user.email],
            subject=cap.config.NOTIFICATIONS.register.subject,
            template=cap.config.NOTIFICATIONS.register.body,
            token=token,
            user=user.to_dict(),
        )

    def register_confirm(self, payload: ObjectDict):
        email = self.auth_code_handler.check(
            self.CACHE_TEMP_KEYS.account_register, payload.token
        )

        try:
            user = self.find_by_email(email)
            user.is_active = True
            user.is_confirmed = True
            self.session.merge(user)
            self.session.commit()
        except SQLAlchemyError as exc:
            cap.logger.exception(exc)
            self.session.rollback()
            abort(httpcode.INTERNAL_SERVER_ERROR)

    def check_user(self, email: str, password: str):
        user = self.find_by_email(email)
        if not user.check_password(password):
            abort(httpcode.BAD_REQUEST, "invalid password")
        return user

    def password_reset(self, data: ObjectDict):
        user = self.check_user(data.email, data.old_password)
        try:
            user.password = data.new_password
            self.session.merge(user)
            self.session.commit()
        except SQLAlchemyError as exc:
            cap.logger.exception(exc)
            self.session.rollback()
            abort(httpcode.INTERNAL_SERVER_ERROR)

    def password_forgot(self, data: ObjectDict):
        user = self.find_by_email(data.email)
        token = self.auth_code_handler.generate(
            self.CACHE_TEMP_KEYS.account_forgot, user
        )
        self.sendmail(
            recipients=[data.email],
            subject=cap.config.NOTIFICATIONS.forgot.subject,
            template=cap.config.NOTIFICATIONS.forgot.body,
            token=token,
            user=user.to_dict(),
        )

    def password_confirm(self, data: ObjectDict):
        email = self.auth_code_handler.check(
            self.CACHE_TEMP_KEYS.account_forgot, data.token
        )
        user = self.find_by_email(email)
        user.password = data.password

        try:
            self.session.merge(user)
            self.session.commit()
        except SQLAlchemyError as exc:
            cap.logger.exception(exc)
            self.session.rollback()
            abort(httpcode.INTERNAL_SERVER_ERROR)


class BaseAccountView(BaseView):
    account_handler: AccountHandler = t.cast(AccountHandler, ExtProxy("account"))
    methods: t.ClassVar[t.Optional[t.Collection[str]]] = [
        HttpMethod.POST,
        HttpMethod.PUT,
    ]

    DEFAULT_SCHEMAS = ObjectDict(
        REGISTER=Fields.object(
            properties={"email": Fields.string, "password": Fields.string}
        ),
        REGISTER_CONFIRM=Fields.object(properties={"token": Fields.string}),
        PASSWORD_RESET=Fields.object(
            properties={
                "email": Fields.string,
                "new_password": Fields.string,
                "old_password": Fields.string,
            }
        ),
        PASSWORD_FORGOT=Fields.object(properties={"email": Fields.string}),
        PASSWORD_CONFIRM=Fields.object(
            properties={"token": Fields.string, "password": Fields.string}
        ),
    )

    def __init__(self):
        self.schemas = cap.config.SCHEMAS.ACCOUNT or self.DEFAULT_SCHEMAS

    def get_payload(self, name: str) -> ObjectDict:
        return PayloadValidator.validate(self.schemas[name])


class RegisterView(BaseAccountView):
    default_view_name = "account_register"
    default_urls = ("/register",)

    def dispatch_request(self, *_, **__):
        if request.method == HttpMethod.POST:
            self.account_handler.register(self.get_payload("REGISTER"))
            return Response.no_content(httpcode.ACCEPTED)

        if request.method == HttpMethod.PUT:
            self.account_handler.register_confirm(self.get_payload("REGISTER_CONFIRM"))
            return Response.no_content()

        return abort(httpcode.METHOD_NOT_ALLOWED)


class PasswordForgotView(BaseAccountView):
    default_view_name = "account_password_forgot"
    default_urls = ("/password/forgot",)

    def dispatch_request(self, *_, **__):
        if request.method == HttpMethod.POST:
            self.account_handler.password_forgot(self.get_payload("PASSWORD_FORGOT"))
            return Response.no_content(httpcode.ACCEPTED)

        if request.method == HttpMethod.PUT:
            self.account_handler.password_confirm(self.get_payload("PASSWORD_CONFIRM"))
            return Response.no_content()

        return abort(httpcode.METHOD_NOT_ALLOWED)


class PasswordResetView(BaseAccountView):
    default_view_name = "account_password_reset"
    default_urls = ("/password/reset",)

    def dispatch_request(self, *_, **__):
        self.account_handler.password_reset(self.get_payload("PASSWORD_RESET"))
        return Response.no_content()



# pylint: disable=redefined-outer-name

import os

import pytest
from flask import Config
from scripts.cli import APP_CONFIG
from vbcore.datastruct import ObjectDict
from vbcore.db.support import SQLASupport

from flaskel.tester import helpers as h, TestClient
from flaskel.tester.helpers import ApiTester

from .helpers import Views

DB_TEST = "test.sqlite"
BASE_DIR = os.path.abspath(os.path.dirname(__file__))
SAMPLE_DATA = os.path.join(BASE_DIR, "..", "resources", "sample.sql")
CONFIG = ObjectDict()

CONFIG.DEBUG = True
CONFIG.TESTING = True
CONFIG.MAIL_DEBUG = True
CONFIG.CACHE_TYPE = "null"
CONFIG.RATELIMIT_ENABLED = False
CONFIG.SQLALCHEMY_ECHO = True
CONFIG.SQLALCHEMY_DATABASE_URI = f"sqlite:///{DB_TEST}"


def load_sample_data():
    SQLASupport.exec_from_file(
        CONFIG.SQLALCHEMY_DATABASE_URI, SAMPLE_DATA, CONFIG.SQLALCHEMY_ECHO
    )


@pytest.fixture(scope="session")
def testapp():
    try:
        os.remove(DB_TEST)
    except OSError as exc:
        print(str(exc))

    config = Config(".")
    config.from_pyfile("{skeleton}/scripts/config.py")
    config.from_mapping(CONFIG)

    extras = dict(
        conf=ObjectDict(**config),
        template_folder="{skeleton}/views/web/templates",
        static_folder="{skeleton}/views/web/static",
        after_create_hook=load_sample_data,
    )

    return TestClient.get_app(**{**APP_CONFIG, **extras})


@pytest.fixture(scope="session")
def test_client(testapp):
    return testapp.test_client()


@pytest.fixture(scope="session")
def config(testapp):
    return testapp.config


@pytest.fixture(scope="session")
def api_tester(test_client):
    return ApiTester(test_client)


@pytest.fixture(scope="session")
def views():
    return Views


@pytest.fixture()
def auth_token(test_client, views):
    def get_access_token(token=None, email=None, password=None, in_query=True):
        if token is not None:
            if in_query is True:
                return f"token={token}"
            return dict(Authorization=f"Bearer {token}")

        credentials = dict(
            email=email or h.config.ADMIN_EMAIL,
            password=password or h.config.ADMIN_PASSWORD,
        )
        tokens = test_client.post(h.url_for(views.access_token), json=credentials)
        if in_query is True:
            return f"token={tokens.json.access_token}"
        return dict(Authorization=f"Bearer {tokens.json.access_token}")

    return get_access_token



import sys
import threading
import traceback
import typing as t
from multiprocessing import cpu_count

from gunicorn.app.base import BaseApplication as WSGIServer

from flaskel import Flaskel
from flaskel.config import config

from .base import BaseApplication


class WSGIGunicorn(BaseApplication, WSGIServer):
    default_log_format = '%(h)s %(l)s %(u)s %(t)s "%(r)s" %(s)s %(b)s "%(f)s" "%(a)s"'

    def __init__(self, app: Flaskel, options: t.Optional[dict] = None):
        BaseApplication.__init__(self, app, options)
        WSGIServer.__init__(self)

    def init(self, parser, opts, args):
        self.options = opts or {}

    def run(self):
        WSGIServer.run(self)

    def set_default_config(self):
        self.cfg.set("preload_app", True)
        self.cfg.set("post_fork", self.post_fork)
        self.cfg.set("pre_fork", self.pre_fork)
        self.cfg.set("pre_exec", self.pre_exec)
        self.cfg.set("when_ready", self.when_ready)
        self.cfg.set("worker_abort", self.worker_abort)
        self.cfg.set("worker_int", self.worker_int)

        host = config("APP_HOST", default="127.0.0.1")
        port = config("APP_PORT", default="5000", cast=int)
        self.cfg.set("bind", config("BIND", default=f"{host}:{port}"))
        self.cfg.set("pidfile", config("PID_FILE", default=".gunicorn.pid"))
        self.cfg.set("proc_name", config("PROC_NAME", default=None))
        self.cfg.set("timeout", config("TIMEOUT", default=30, cast=int))
        self.cfg.set("backlog", config("BACKLOG", default=2048, cast=int))
        self.cfg.set("keepalive", config("KEEPALIVE", default=3, cast=int))

        nw = 1 + 2 * cpu_count()
        worker_class = config("WORKER_CLASS", default="sync")
        self.cfg.set("worker_class", worker_class)

        self.cfg.set(
            "workers",
            config("WORKERS", default=nw if worker_class == "sync" else 1, cast=int),
        )
        self.cfg.set(
            "threads",
            config("THREADS", default=nw if worker_class == "gthread" else 1, cast=int),
        )

        self.cfg.set(
            "worker_connections", config("WORKER_CONNECTIONS", default=1000, cast=int)
        )

        self.cfg.set("errorlog", "-")
        self.cfg.set("accesslog", "-")
        self.cfg.set("loglevel", config("LOG_LEVEL", default="info").lower())
        self.cfg.set("access_log_format", self.default_log_format)

    def load_config(self):
        self.set_default_config()

        config_items = (
            (key, value)
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        )
        for key, value in config_items:
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application

    @staticmethod
    def post_fork(server, worker):
        server.log.info("Worker spawned (pid: %s)", worker.pid)

    @staticmethod
    def pre_fork(server, worker):
        server.log.info("pre fork (pid: %s)", worker.pid)

    @staticmethod
    def pre_exec(server):
        server.log.info("Forked child, re-executing.")

    @staticmethod
    def when_ready(server):
        server.log.info("Server is ready. Spawning workers")

    @staticmethod
    def worker_abort(worker):
        worker.log.info("worker received SIGABRT signal")

    @staticmethod
    def nworkers_changed(server, new_value, old_value):
        server.log.info(f"number of workers changed from {old_value} to {new_value}")

    @staticmethod
    def worker_int(worker):
        worker.log.info("worker received INT or QUIT signal")

        class DumpTraceback:
            def __str__(self):
                code = []
                id2name = {th.ident: th.name for th in threading.enumerate()}
                # noinspection PyUnresolvedReferences, PyProtectedMember
                for thread_id, stack in sys._current_frames().items():
                    thread_name = id2name.get(thread_id, "")
                    code.append(f"\n# Thread: {thread_name}({thread_id})")
                    for filename, lineno, name, line in traceback.extract_stack(stack):
                        code.append(f"File: '{filename}', line {lineno}, in {name}")
                        if line:
                            code.append(f"\t{line.strip()}")
                return "\n".join(code)

        worker.log.warning("%s", DumpTraceback())



import typing as t

from vbcore.importer import ImporterError, ImporterFactory

from .base import BaseApplication

DEFAULT_WSGI_SERVERS: t.Dict[str, str] = {
    "builtin": "flaskel.wsgi.base:WSGIBuiltin",
    "gunicorn": "flaskel.wsgi.wsgi_gunicorn:WSGIGunicorn",
    "gevent": "flaskel.wsgi.wsgi_gevent:WSGIGevent",
    "tornado": "flaskel.wsgi.wsgi_tornado.WSGITornado",
    "twisted": "flaskel.wsgi.wsgi_twisted:WSGITwisted",
    "waitress": "flaskel.wsgi.wsgi_waitress:WSGIWaitress",
}


class WSGIFactory(ImporterFactory):
    def get_class(self, name: str, *args, **kwargs) -> t.Type[BaseApplication]:
        try:
            return super().get_class(name, *args, **kwargs)
        except ImporterError as exc:
            raise ImportError(f"missing '{name}' dependency") from exc



import ipaddress
import typing as t
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from functools import partial

from redis import Redis
from vbcore.datastruct import ExpiringCache, LRUCache
from vbcore.date_helper import Day
from vbcore.http import httpcode

from flaskel import abort, client_redis, flaskel

cap = flaskel.cap

IpType = t.Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
NetType = t.Union[ipaddress.IPv4Network, ipaddress.IPv6Network]


class IBanRepo(ABC):
    def __init__(self, *_, key_prefix: str = "", separator: str = "/", **__):
        self.separator = separator
        self.key_prefix = key_prefix

    def prepare_key(self, key: str) -> str:
        return f"{self.key_prefix}{self.separator}{key}"

    @abstractmethod
    def attempts(self, key: str) -> int:
        pass  # pragma: no cover

    @abstractmethod
    def ban(self, key: str, ttl: t.Optional[int] = None) -> int:
        pass  # pragma: no cover

    @abstractmethod
    def unban(self, key: str):
        pass  # pragma: no cover


class BanRepoRedis(IBanRepo):
    def __init__(self, client: Redis, **kwargs):
        super().__init__(**kwargs)
        self.client = client

    def attempts(self, key: str) -> int:
        value = self.client.get(self.prepare_key(key))
        return int(value) if value else 0

    def ban(self, key: str, ttl: t.Optional[int] = None) -> int:
        cache_key = self.prepare_key(key)
        pipe = self.client.pipeline()
        pipe.incr(cache_key)
        if ttl is not None:
            pipe.expire(cache_key, ttl)
        result = pipe.execute()
        return result[0]

    def unban(self, key: str):
        self.client.delete(self.prepare_key(key))


class BanRepoLocal(IBanRepo):
    def __init__(
        self,
        client_class: t.Type[LRUCache] = LRUCache,
        maxsize: int = 1000,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.client = client_class(maxsize)

    def get(self, key: str) -> t.Tuple[int, t.Optional[datetime]]:
        attempts, set_time = self.client.get(self.prepare_key(key)) or (0, None)
        if set_time and set_time < datetime.now():
            self.unban(key)
            return 0, None
        return attempts, set_time

    def attempts(self, key: str) -> int:
        return self.get(key)[0]

    def ban(self, key: str, ttl: t.Optional[int] = None) -> int:
        attempts, set_time = self.get(key)
        set_time = (
            None
            if ttl is None
            else (set_time or datetime.now()) + timedelta(seconds=ttl)
        )
        self.client.set(self.prepare_key(key), (attempts + 1, set_time))
        return attempts + 1

    def unban(self, key: str):
        try:
            del self.client[self.prepare_key(key)]
        except KeyError:
            pass


class IpBanService:
    def __init__(
        self,
        repo_class: t.Type[IBanRepo],
        key_prefix: str,
        max_attempts: t.Optional[int] = None,
        default_ttl: t.Optional[int] = None,
        **kwargs,
    ):
        self.default_ttl = default_ttl
        self.max_attempts = max_attempts
        self.white_list: t.Set[IpType] = set()
        self.net_white_list: t.Set[NetType] = set()
        self.repo: IBanRepo = repo_class(key_prefix=key_prefix, **kwargs)
        self.black_list: IBanRepo = repo_class(
            key_prefix=self.repo.prepare_key("blacklist"), **kwargs
        )

    def load_whitelist(self, ip: t.Iterable[str] = (), net: t.Iterable[str] = ()):
        self.white_list.update(ipaddress.ip_address(i) for i in ip)
        self.net_white_list.update(ipaddress.ip_network(n) for n in net)

    def is_whitelisted(self, ip: str) -> bool:
        ip_addr = ipaddress.ip_address(ip)
        if ip_addr in self.white_list:
            return True
        for net in self.net_white_list:
            if ip_addr in net:
                return True
        return False

    def is_blacklisted(self, ip: str) -> bool:
        return bool(self.black_list.attempts(ip))

    def add_blacklist(self, ip: str):
        self.black_list.ban(ip, ttl=None)  # never expires

    def ban(
        self, ip: str, ttl: t.Optional[int] = None, permanent: bool = False
    ) -> t.Optional[int]:
        if permanent is True:
            return self.black_list.ban(ip)

        if self.is_whitelisted(ip) or self.is_blacklisted(ip):
            return None
        return self.repo.ban(ip, ttl or self.default_ttl)

    def is_banned(self, ip: str) -> bool:
        if self.is_whitelisted(ip):
            return False
        if self.is_blacklisted(ip):
            return True

        return self.repo.attempts(ip) >= self.max_attempts


class FlaskIPBan:
    def __init__(self, app=None, **kwargs) -> None:
        self.service: t.Optional[IpBanService] = None
        self.backends: t.Dict[str, t.Type[IBanRepo]] = {}

        self.register_backend("local", BanRepoLocal, client_class=ExpiringCache)
        self.register_backend("redis", BanRepoRedis, client=client_redis)

        if app is not None:
            self.init_app(app, **kwargs)

    @classmethod
    def set_default_config(cls, app):
        app.config.setdefault("IPBAN_ENABLED", True)
        app.config.setdefault("IPBAN_KEY_PREFIX", app.config.APP_NAME)
        app.config.setdefault("IPBAN_KEY_SEP", "/")
        app.config.setdefault("IPBAN_BACKEND", "local")
        app.config.setdefault("IPBAN_BACKEND_OPTS", {})
        app.config.setdefault("IPBAN_COUNT", 20)
        app.config.setdefault("IPBAN_SECONDS", Day.seconds)
        app.config.setdefault("IPBAN_NET_WHITELIST", ["127.0.0.0/8"])
        app.config.setdefault("IPBAN_IP_WHITELIST", ["127.0.0.1"])
        app.config.setdefault("IPBAN_STATUS_CODE", httpcode.FORBIDDEN)
        app.config.setdefault(
            "IPBAN_CHECK_CODES",
            (
                httpcode.NOT_FOUND,
                httpcode.METHOD_NOT_ALLOWED,
                httpcode.NOT_IMPLEMENTED,
            ),
        )

    def init_app(self, app, service_class: t.Type[IpBanService] = IpBanService):
        self.set_default_config(app)
        if app.config.IPBAN_ENABLED is True:
            app.after_request(self.after_request_hook)
            app.before_request(self.before_request_hook)
            self.service = self.service_factory(service_class, app.config)
            self.service.load_whitelist(
                ip=app.config.IPBAN_IP_WHITELIST, net=app.config.IPBAN_NET_WHITELIST
            )

        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["ipban"] = self

    def service_factory(self, service_class: t.Type[IpBanService], config):
        return service_class(
            self.backends[config.IPBAN_BACKEND],
            max_attempts=config.IPBAN_COUNT,
            default_ttl=config.IPBAN_SECONDS,
            key_prefix=config.IPBAN_KEY_PREFIX,
            separator=config.IPBAN_KEY_SEP,
            **config.IPBAN_BACKEND_OPTS,
        )

    def register_backend(
        self, name: str, repo_class: t.Type[IBanRepo], *args, **kwargs
    ):
        backend = partial(repo_class, *args, **kwargs)
        self.backends[name] = t.cast(t.Type[IBanRepo], backend)

    @classmethod
    def get_ip(cls) -> str:
        return flaskel.request.remote_addr

    def after_request_hook(self, response):
        if response.status_code in cap.config.IPBAN_CHECK_CODES:
            ip = self.get_ip()
            attempts = self.service.ban(ip)
            if attempts:
                cap.logger.info("%s added to ban list with attempts: %s", ip, attempts)
        return response

    def before_request_hook(self):
        ip = self.get_ip()
        if self.service.is_banned(ip):
            cap.logger.warning("%s is banned", ip)
            abort(cap.config.IPBAN_STATUS_CODE)


ipban = FlaskIPBan()



import os
import typing as t

import click
from vbcore import yaml
from vbcore.datastruct import ObjectDict
from vbcore.misc import parse_value

from .builder import AppBuilder
from .config import config as configurator
from .wsgi import BaseApplication, DEFAULT_WSGI_SERVERS, WSGIFactory


def option_as_dict(ctx, param, value):
    _ = ctx, param
    ret = {}
    for opt in value:
        k, v = opt.split("=", 2)
        ret.update({k: parse_value(v)})
    return ret


class Server:
    opt_config_attr = dict(default=None, help="app yaml/json configuration file")
    opt_log_config_attr = dict(
        default=None, help="alternative log yaml/json configuration file"
    )
    opt_bing_attr = dict(
        default="127.0.0.1:5000", help="address to bind", show_default=True
    )
    opt_debug_attr = dict(
        is_flag=True, flag_value=True, default=False, help="enable debug mode"
    )
    opt_wsgi_server_attr = dict(
        default=None,
        type=click.Choice(list(DEFAULT_WSGI_SERVERS.keys()), case_sensitive=False),
        help="name of wsgi server to use",
    )
    option_wsgi_config_attr = dict(
        default={},
        multiple=True,
        callback=option_as_dict,
        metavar="KEY=VAL",
        help="wsgi configuration",
    )
    option_app_config_attr = dict(
        default={},
        multiple=True,
        callback=option_as_dict,
        metavar="KEY=VAL",
        help="app configuration",
    )

    def __init__(
        self,
        app_factory: t.Optional[AppBuilder] = None,
        wsgi_factory: t.Optional[WSGIFactory] = None,
    ):
        self._app_factory = app_factory or AppBuilder()
        self._wsgi_factory = wsgi_factory or WSGIFactory(
            base_class=BaseApplication, classes=DEFAULT_WSGI_SERVERS
        )

    def register_options(self, func: t.Callable) -> t.Callable:
        @click.command()
        @click.option("-b", "--bind", **self.opt_bing_attr)
        @click.option("-d", "--debug", **self.opt_debug_attr)
        @click.option("-c", "--config", **self.opt_config_attr)
        @click.option("-l", "--log-config", **self.opt_log_config_attr)
        @click.option("-w", "--wsgi-server", **self.opt_wsgi_server_attr)
        @click.option("-D", "--wsgi-config", **self.option_wsgi_config_attr)
        @click.option("-W", "--app-config", **self.option_app_config_attr)
        def callable_options(*args, **kwargs):
            return func(*args, **kwargs)

        return callable_options

    def run_from_cli(self, **kwargs):
        runner = self.runner(**kwargs)
        return runner()

    def runner(self, **kwargs) -> t.Callable:
        @self.register_options
        def forever(**params):
            self.serve_forever(**params, **kwargs)

        return forever

    @staticmethod
    def prepare_config(
        filename: str,
        debug: bool = False,
        bind: t.Optional[str] = None,
        log_config: t.Optional[str] = None,
    ):
        default_env = "development" if debug else "production"
        env = configurator("FLASK_ENV", default=default_env)

        if filename is not None:
            config = yaml.load_yaml_file(filename)
            if not config.app.FLASK_ENV:
                config.app.FLASK_ENV = env
        else:
            config = ObjectDict(
                app={"DEBUG": debug, "FLASK_ENV": env},
                wsgi={"bind": bind, "debug": debug},
            )

        if log_config is not None:
            config.app.LOG_FILE_CONF = log_config

        # debug flag enabled overrides config file
        if debug is True:
            config.app.DEBUG = True
            os.environ["FLASK_DEBUG"] = "1"
        if bind is not None:
            config.wsgi.bind = bind

        os.environ["FLASK_ENV"] = config.app.FLASK_ENV
        return config

    def serve_forever(
        self,
        config: t.Optional[str] = None,
        log_config: t.Optional[str] = None,
        bind: t.Optional[str] = None,
        wsgi_class: t.Type[BaseApplication] = None,
        wsgi_server: t.Optional[str] = None,
        wsgi_config: t.Optional[dict] = None,
        app_config: t.Optional[dict] = None,
        debug: bool = False,
    ):
        """

        :param config: app and wsgi configuration file
        :param log_config: log configuration file
        :param bind: address to bind
        :param debug: enable debug mode
        :param wsgi_class: optional a custom subclass of BaseApplication
        :param wsgi_server: wsgi server chosen
        :param wsgi_config: wsgi configuration
        :param app_config: app configuration
        :return: never returns
        """
        config_data = self.prepare_config(config, debug, bind, log_config)
        wsgi_server = wsgi_server or config_data.wsgi_server
        conf_wsgi = config_data.get("wsgi_opts") or config_data.get("wsgi") or {}
        wsgi_config = {**conf_wsgi, **(wsgi_config or {})}
        app_config = {**(config_data.get("app") or {}), **(app_config or {})}

        if wsgi_server:
            wsgi_class = self._wsgi_factory.get_class(wsgi_server)
        elif wsgi_class is None:
            wsgi_class = self._wsgi_factory.get_class("builtin")

        app = self._app_factory.create(app_config)
        wsgi = wsgi_class(app, options=wsgi_config)
        wsgi.run()  # run forever



import socket
import typing as t

from flask_mail import Attachment, Mail, Message

AddressType = t.Union[str, t.Tuple[str, str]]


class ClientMail(Mail):
    def init_app(self, app):
        self.app = app
        super().init_app(app)
        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["client_mail"] = self

    def sendmail(
        self,
        subject: str,
        html: t.Optional[str] = None,
        body: t.Optional[str] = None,
        recipients: t.Optional[t.List[AddressType]] = None,
        reply_to: t.Optional[AddressType] = None,
        cc: t.Optional[t.List[AddressType]] = None,
        bcc: t.Optional[t.List[AddressType]] = None,
        sender: t.Optional[AddressType] = None,
        attachments: t.Optional[t.List[Attachment]] = None,
        **kwargs,
    ) -> str:
        sender = sender or self.app.config.MAIL_DEFAULT_SENDER
        destination = recipients or [self.app.config.MAIL_RECIPIENT]
        message = Message(
            subject=subject,
            html=html,
            body=body,
            recipients=destination,
            reply_to=reply_to,
            cc=cc,
            bcc=bcc,
            sender=sender,
            attachments=attachments,
            **kwargs,
        )

        timeout = socket.getdefaulttimeout()
        socket.setdefaulttimeout(self.app.config.MAIL_TIMEOUT or 60)
        self.send(message)
        socket.setdefaulttimeout(timeout)

        self.app.logger.info(
            "sent email %s from %s to %s", message.msgId, sender, destination
        )
        return message.msgId



import inspect
import sys

import flask
from vbcore.datastruct import ObjectDict
from vbcore.http import httpcode
from werkzeug import exceptions

from flaskel.utils.datastruct import ConfigProxy


class HTTPExceptionMixin:
    description = ""
    config = ConfigProxy("HTTP_EXCEPTIONS")
    lang_key = ConfigProxy("HTTP_EXCEPTIONS.LANG_KEY")
    lang_default = ConfigProxy("HTTP_EXCEPTIONS.LANG_DEFAULT")

    # noinspection PyUnusedLocal
    def get_description(self, environ=None, scope=None) -> str:  # pylint: disable=W0613
        lang = self.config.get(flask.g.get(self.lang_key.get() or "lang"))
        lang = lang or self.config.get(self.lang_default.get()) or {}
        desc = lang.get(self.description)
        desc = desc or self.config.get(self.description)
        desc = desc or self.description.lower().replace("_", " ")
        return desc or "unhandled error"


class BadRequest(HTTPExceptionMixin, exceptions.BadRequest):
    code = httpcode.BAD_REQUEST
    description = "BAD_REQUEST"


class Unauthorized(HTTPExceptionMixin, exceptions.Unauthorized):
    code = httpcode.UNAUTHORIZED
    description = "UNAUTHORIZED"


class PaymentRequired(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.PAYMENT_REQUIRED
    description = "PAYMENT_REQUIRED"


class Forbidden(HTTPExceptionMixin, exceptions.Forbidden):
    code = httpcode.FORBIDDEN
    description = "FORBIDDEN"


class NotFound(HTTPExceptionMixin, exceptions.NotFound):
    code = httpcode.NOT_FOUND
    description = "NOT_FOUND"


class MethodNotAllowed(HTTPExceptionMixin, exceptions.MethodNotAllowed):
    code = httpcode.METHOD_NOT_ALLOWED
    description = "METHOD_NOT_ALLOWED"


class NotAcceptable(HTTPExceptionMixin, exceptions.NotAcceptable):
    code = httpcode.NOT_ACCEPTABLE
    description = "NOT_ACCEPTABLE"


class ProxyAuthenticationRequired(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.PROXY_AUTHENTICATION_REQUIRED
    description = "PROXY_AUTHENTICATION_REQUIRED"


class RequestTimeout(HTTPExceptionMixin, exceptions.RequestTimeout):
    code = httpcode.REQUEST_TIMEOUT
    description = "REQUEST_TIMEOUT"


class Conflict(HTTPExceptionMixin, exceptions.Conflict):
    code = httpcode.CONFLICT
    description = "CONFLICT"


class Gone(HTTPExceptionMixin, exceptions.Gone):
    code = httpcode.GONE
    description = "GONE"


class LengthRequired(HTTPExceptionMixin, exceptions.LengthRequired):
    code = httpcode.LENGTH_REQUIRED
    description = "LENGTH_REQUIRED"


class PreconditionFailed(HTTPExceptionMixin, exceptions.PreconditionFailed):
    code = httpcode.PRECONDITION_FAILED
    description = "PRECONDITION_FAILED"


class RequestEntityTooLarge(HTTPExceptionMixin, exceptions.RequestEntityTooLarge):
    code = httpcode.REQUEST_ENTITY_TOO_LARGE
    description = "REQUEST_ENTITY_TOO_LARGE"


class RequestURITooLarge(HTTPExceptionMixin, exceptions.RequestURITooLarge):
    code = httpcode.REQUEST_URI_TOO_LONG
    description = "REQUEST_URI_TOO_LONG"


class UnsupportedMediaType(HTTPExceptionMixin, exceptions.UnsupportedMediaType):
    code = httpcode.UNSUPPORTED_MEDIA_TYPE
    description = "UNSUPPORTED_MEDIA_TYPE"


class RequestedRangeNotSatisfiable(
    HTTPExceptionMixin, exceptions.RequestedRangeNotSatisfiable
):
    code = httpcode.RANGE_NOT_SATISFIABLE
    description = "RANGE_NOT_SATISFIABLE"


class ExpectationFailed(HTTPExceptionMixin, exceptions.ExpectationFailed):
    code = httpcode.EXPECTATION_FAILED
    description = "EXPECTATION_FAILED"


class MisdirectedRequest(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.MISDIRECTED_REQUEST
    description = "MISDIRECTED_REQUEST"


class UnprocessableEntity(HTTPExceptionMixin, exceptions.UnprocessableEntity):
    code = httpcode.UNPROCESSABLE_ENTITY
    description = "UNPROCESSABLE_ENTITY"


class Locked(HTTPExceptionMixin, exceptions.Locked):
    code = httpcode.LOCKED
    description = "LOCKED"


class FailedDependency(HTTPExceptionMixin, exceptions.FailedDependency):
    code = httpcode.FAILED_DEPENDENCY
    description = "FAILED_DEPENDENCY"


class TooEarly(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.TOO_EARLY
    description = "TOO_EARLY"


class UpgradeRequired(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.UPGRADE_REQUIRED
    description = "UPGRADE_REQUIRED"


class PreconditionRequired(HTTPExceptionMixin, exceptions.PreconditionRequired):
    code = httpcode.PRECONDITION_REQUIRED
    description = "PRECONDITION_REQUIRED"


class TooManyRequests(HTTPExceptionMixin, exceptions.TooManyRequests):
    code = httpcode.TOO_MANY_REQUESTS
    description = "TOO_MANY_REQUESTS"


class RequestHeaderFieldsTooLarge(
    HTTPExceptionMixin, exceptions.RequestHeaderFieldsTooLarge
):
    code = httpcode.REQUEST_HEADER_FIELDS_TOO_LARGE
    description = "REQUEST_HEADER_FIELDS_TOO_LARGE"


class NoResponse(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.NO_RESPONSE
    description = "NO_RESPONSE"


class UnavailableForLegalReasons(
    HTTPExceptionMixin, exceptions.UnavailableForLegalReasons
):
    code = httpcode.UNAVAILABLE_FOR_LEGAL_REASON
    description = "UNAVAILABLE_FOR_LEGAL_REASON"


class InternalServerError(HTTPExceptionMixin, exceptions.InternalServerError):
    code = httpcode.INTERNAL_SERVER_ERROR
    description = "INTERNAL_SERVER_ERROR"


# noinspection PyShadowingBuiltins
# pylint: disable=redefined-builtin
class NotImplemented(HTTPExceptionMixin, exceptions.NotImplemented):
    code = httpcode.NOT_IMPLEMENTED
    description = "NOT_IMPLEMENTED"


class BadGateway(HTTPExceptionMixin, exceptions.BadGateway):
    code = httpcode.BAD_GATEWAY
    description = "BAD_GATEWAY"


class ServiceUnavailable(HTTPExceptionMixin, exceptions.ServiceUnavailable):
    code = httpcode.SERVICE_UNAVAILABLE
    description = "SERVICE_UNAVAILABLE"


class GatewayTimeout(HTTPExceptionMixin, exceptions.GatewayTimeout):
    code = httpcode.GATEWAY_TIMEOUT
    description = "GATEWAY_TIMEOUT"


class HTTPVersionNotSupported(HTTPExceptionMixin, exceptions.HTTPVersionNotSupported):
    code = httpcode.HTTP_VERSION_NOT_SUPPORTED
    description = "HTTP_VERSION_NOT_SUPPORTED"


class InsufficientStorage(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.INSUFFICIENT_STORAGE
    description = "INSUFFICIENT_STORAGE"


class LoopDetected(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.LOOP_DETECTED
    description = "LOOP_DETECTED"


class BandwidthLimit(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.BANDWIDTH_LIMIT
    description = "BANDWIDTH_LIMIT"


class NotExtended(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.NOT_EXTENDED
    description = "NOT_EXTENDED"


class NetworkAuthenticationRequired(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.NETWORK_AUTHENTICATION_REQUIRED
    description = "NETWORK_AUTHENTICATION_REQUIRED"


class UnknownError(HTTPExceptionMixin, exceptions.HTTPException):
    code = httpcode.UNKNOWN_ERROR
    description = "UNKNOWN_ERROR"


class RPCError(Exception):
    def __init__(self, code, message, data=None):
        super().__init__(code, message)
        self.code = code
        self.message = message
        self.data = data

    def as_dict(self):
        return ObjectDict(code=self.code, message=self.message, data=self.data)


class RPCParseError(RPCError):
    def __init__(self, message="Invalid JSON was received by the server", data=None):
        super().__init__(-32700, message, data)


class RPCInvalidRequest(RPCError):
    def __init__(
        self,
        message="The JSON sent is not a valid Request object",
        data=None,
        req_id=None,
    ):
        super().__init__(-32600, message, data)
        self.req_id = req_id


class RPCMethodNotFound(RPCError):
    def __init__(
        self, message="The method does not exist or is not available", data=None
    ):
        super().__init__(-32601, message, data)


class RPCInvalidParams(RPCError):
    def __init__(self, message="Invalid method parameter(s)", data=None):
        super().__init__(-32602, message, data)


class RPCInternalError(RPCError):
    def __init__(self, message="Internal JSON-RPC error", data=None):
        super().__init__(-32603, message, data)


def rpc_error_to_httpcode(error_code: int) -> int:
    if error_code == RPCParseError().code:
        return httpcode.BAD_REQUEST
    if error_code == RPCInvalidRequest().code:
        return httpcode.BAD_REQUEST
    if error_code == RPCMethodNotFound().code:
        return httpcode.NOT_FOUND
    if error_code == RPCInvalidParams().code:
        return httpcode.UNPROCESSABLE_ENTITY

    return httpcode.INTERNAL_SERVER_ERROR


# until a better way is found
errors = inspect.getmembers(
    sys.modules[__name__],
    lambda c: inspect.isclass(c) and issubclass(c, exceptions.HTTPException),
)
abort = exceptions.Aborter(extra={e.code: e for _, e in errors})
setattr(flask, "abort", abort)



import typing as t

from flask_sqlalchemy import Model
from vbcore.datastruct import ObjectDict


class SQLAModel(Model):
    __table__ = None

    def columns(self) -> t.List[str]:
        if self.__table__ is not None:
            return list(self.__table__.columns.keys())
        return []

    @classmethod
    def get_one(
        cls, *args, raise_not_found: bool = True, to_dict: bool = True, **kwargs
    ):
        res = cls.query.filter(*args).filter_by(**kwargs)

        if raise_not_found:
            res = res.first_or_404()
        else:
            res = res.first()
            if res is None:
                return None

        if to_dict is True:
            return res.to_dict()
        return res

    @classmethod
    def prepare_collection_filters(cls, params: dict) -> list:
        # stub method
        _ = params
        return []

    @classmethod
    def query_collection(cls, *_, params: t.Optional[dict] = None, **kwargs):
        filters = cls.prepare_collection_filters(params or {})
        return cls.query.filter(*filters).filter_by(**kwargs)

    @classmethod
    def get_list(
        cls,
        *args,
        to_dict: bool = True,
        restricted: bool = False,
        order_by: t.Optional[t.Tuple] = None,
        page: t.Optional[int] = None,
        page_size: t.Optional[int] = None,
        max_per_page: t.Optional[int] = None,
        **kwargs,
    ):
        q = cls.query_collection(*args, **kwargs)

        if order_by is not None:
            q = q.order_by(*order_by)

        if page or page_size:
            q = q.paginate(
                page=page,
                per_page=page_size,
                error_out=False,
                max_per_page=max_per_page,
            )
            res = q.items
            if to_dict is False:
                return q
        else:
            res = q.all()

        if to_dict is True:
            return (r.to_dict(restricted) for r in res)
        return res

    def update(self, attributes: dict):
        for attr, val in attributes.items():
            if attr in self.columns():
                setattr(self, attr, val)

        return self

    def to_dict(self, restricted: bool = False) -> dict:
        _ = restricted
        # noinspection PyUnresolvedReferences
        cols = self.columns()  # type: ignore
        return ObjectDict(**{c: getattr(self, c, None) for c in cols})



from flaskel.ext import default, limit
from flaskel.ext.auth import token_auth
from flaskel.ext.mongo import FlaskMongoDB
from flaskel.ext.redis import FlaskRedis
from flaskel.ext.sendmail import ClientMail
from flaskel.extra.mobile_support import MobileVersionCompatibility, RedisStore
from flaskel.extra.payments.stripe import PaymentHandler

from .auth import account_handler
from .database import database

ipban = limit.FlaskIPBan()
client_mail = ClientMail()
client_redis = FlaskRedis()
client_mongo = FlaskMongoDB()
scheduler = default.Scheduler()
payment_handler = PaymentHandler()
mobile_version = MobileVersionCompatibility()


class OPTS:
    cors = dict(
        resources={r"/*": {"origins": "*"}},
    )
    mobile_version = dict(
        store=RedisStore(client_redis),
    )
    errors = dict(
        dispatcher="subdomain",
        response=default.builder.on_accept(strict=False),
    )


EXTENSIONS = {
    "cfremote": (default.cfremote,),  # MUST be the first
    "logger": (default.logger,),  # MUST be the second
    "argon2": (default.argon2,),
    "builder": (default.builder,),
    "cache": (default.caching,),
    "cors": (default.cors, OPTS.cors),
    "database": (database,),
    "date_helper": (default.date_helper,),
    "errors": (default.error_handler, OPTS.errors),
    "health_checks": (default.health_checks,),
    "ip_ban": (ipban,),
    "token_auth": (token_auth,),
    "limiter": (limit.limiter,),
    "mobile_version": (mobile_version, OPTS.mobile_version),
    "mongo": (client_mongo,),
    "redis": (client_redis,),
    "scheduler": (scheduler,),
    "sendmail": (client_mail,),
    "stripe": (payment_handler,),
    "template": (default.template,),
    "useragent": (default.useragent,),
    "account": (account_handler,),
}



import typing as t
from functools import partial

import stripe
from stripe.error import StripeError
from vbcore.datastruct import ObjectDict

from flaskel import cap, request

CallbackType = t.Callable[[t.Dict[str, t.Any]], t.Any]


class PaymentHandler:
    handler = stripe
    payment_intent_ok: str = "payment_intent.succeeded"
    payment_intent_ko: str = "payment_intent.payment_failed"

    def __init__(self, app=None, **kwargs) -> None:
        self._error: CallbackType = partial(self._not_registered, "on_error")
        self._success: CallbackType = partial(self._not_registered, "on_success")

        if app is not None:
            self.init_app(app, **kwargs)

    @staticmethod
    def _not_registered(callback_type: str, *_):
        cap.logger.warning("no callback registered for: %s", callback_type)

    @staticmethod
    def check_prerequisites(app):
        assert app.config.STRIPE_SECRET_KEY is not None, "missing STRIPE_SECRET_KEY"
        assert app.config.STRIPE_PUBLIC_KEY is not None, "missing STRIPE_PUBLIC_KEY"

    def init_app(self, app, name: t.Optional[str] = None, **kwargs):
        self.check_prerequisites(app)
        app.config.setdefault("STRIPE_DEBUG", False)
        app.config.setdefault("STRIPE_DEFAULT_CURRENCY", "eur")
        app.config.setdefault("STRIPE_API_VERSION", "2020-08-27")

        self.handler.enable_telemetry = False
        self.handler.api_key = app.config.STRIPE_SECRET_KEY
        self.handler.api_version = app.config.STRIPE_API_VERSION
        self.handler.log = "debug" if app.config.STRIPE_DEBUG else "info"
        self.handler.set_app_info(name or app.config.APP_NAME, **kwargs)

        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["stripe"] = self

    @staticmethod
    def prepare_amount(value):
        return int(value * 100)

    def create_payment_intent(
        self, amount: t.Union[float, int], currency: t.Optional[str] = None, **kwargs
    ) -> ObjectDict:
        return ObjectDict(
            **self.handler.PaymentIntent.create(
                amount=self.prepare_amount(amount),
                currency=currency or cap.config.STRIPE_DEFAULT_CURRENCY,
                **kwargs,
            )
        )

    def create_event(self, data: bytes, raise_exc: bool = True) -> ObjectDict:
        webhook_secret = cap.config.STRIPE_WEBHOOK_SECRET
        signature = request.headers.get("stripe-signature")
        try:
            return ObjectDict(
                **self.handler.Webhook.construct_event(
                    payload=data, sig_header=signature, secret=webhook_secret
                )
            )
        except StripeError as exc:
            cap.logger.exception(exc)
            if not raise_exc:
                return ObjectDict()
            raise

    def webhook_handler(self) -> t.Tuple[t.Optional[bool], ObjectDict]:
        data = self.create_event(request.data)
        event_type = data.type

        if event_type == self.payment_intent_ok:
            self._success(data)
            return True, data

        if event_type == self.payment_intent_ko:
            self._error(data)
            return True, data

        cap.logger.warning("unhandled event type received: %s\n%s", event_type, data)
        return None, data

    def on_success(self, callback: CallbackType) -> CallbackType:
        self._success = callback
        return callback

    def on_error(self, callback: CallbackType) -> CallbackType:
        self._error = callback
        return callback



from functools import partial

import flask
from vbcore.http.useragent import UserAgent as UserAgentParser

from flaskel.flaskel import request


class UserAgent:
    def __init__(self, app=None, **kwargs):
        self._parser = None

        if app is not None:
            self.init_app(app, **kwargs)

    def init_app(self, app, parser_class=UserAgentParser):
        self._parser = parser_class
        app.config.setdefault("USER_AGENT_AUTO_PARSE", False)

        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["useragent"] = self

        app.before_request_funcs.setdefault(None, []).append(
            partial(self.before_request_hook, app)
        )

    def before_request_hook(self, app):
        flask.g.user_agent = self._parser  # pylint: disable=assigning-non-slot
        if app.config.USER_AGENT_AUTO_PARSE:
            ua_str = request.user_agent.string
            # pylint: disable=assigning-non-slot
            flask.g.user_agent = self._parser.parse(ua_str)



import typing as t

from sqlalchemy.exc import SQLAlchemyError
from vbcore.db.exceptions import DBError
from vbcore.db.support import SQLASupport
from vbcore.http import httpcode, HttpMethod

from flaskel import abort, cap, db_session, PayloadValidator, Response, webargs
from flaskel.ext.default import builder

from ..utils.datastruct import Pagination
from .base import BaseView, Resource, UrlsType


class CatalogResource(Resource):
    pagination_enabled: bool = True

    methods_collection = [
        HttpMethod.GET,
    ]
    methods_resource = [
        HttpMethod.GET,
    ]
    methods_subresource = [
        HttpMethod.GET,
    ]

    def __init__(self, model):
        self._model = model

    def on_get(self, res_id, *_, model=None, **kwargs):
        """
        Get resource info

        :param res_id: resource identifier (primary key value)
        :param model: alternative SQLAlchemy model class
        :param kwargs: extra query filters
        :return:
        """
        model = model or self._model
        return model.get_one(id=res_id, **kwargs)

    def on_collection(self, *_, params=None, model=None, **kwargs):
        """
        Resource collection paginated and sorted

        :param params: parameters (usually from query string)
        :param model: alternative SqlAlchemy model class
        :param kwargs: extra query filters
        :return:
        """
        page = size = None
        model = model or self._model
        max_size = cap.config.PAGINATION_MAX_PAGE_SIZE
        order_by = getattr(model, "order_by", None)

        if params is None:
            params = webargs.paginate()

        if self.pagination_enabled is True:
            page = params.get("page")
            size = params.get("page_size")
            size = max(size, max_size or 0) if size else max_size

        response = model.get_list(
            to_dict=not self.pagination_enabled,
            order_by=order_by,
            page=page,
            page_size=size,
            max_per_page=max_size,
            params=params,
            **kwargs,
        )

        if self.pagination_enabled is True:
            return self.response_paginated(
                response,
                restricted=not params.get("related", False),
            )
        return response

    @classmethod
    def response_paginated(cls, res, **kwargs):
        """
        Prepare the paginated response for resource collection

        :param res: list of sqlalchemy models
        :return:
        """
        if isinstance(res, list):
            return [r.to_dict(**kwargs) for r in res]

        headers = Response.pagination_headers(
            res.total, Pagination(page=res.page, page_size=res.per_page)
        )

        return (
            [r.to_dict(**kwargs) for r in res.items],
            (httpcode.PARTIAL_CONTENT if res.has_next else httpcode.SUCCESS),
            headers,
        )


class Restful(CatalogResource):
    post_schema: t.Any = None
    put_schema: t.Any = None
    support_class: t.Type[SQLASupport] = SQLASupport
    validator: t.Type[PayloadValidator] = PayloadValidator

    methods_subresource = [
        HttpMethod.GET,
        HttpMethod.POST,
    ]
    methods_collection = [
        HttpMethod.GET,
        HttpMethod.POST,
        HttpMethod.PUT,
    ]
    methods_resource = [
        HttpMethod.GET,
        HttpMethod.PUT,
        HttpMethod.DELETE,
    ]

    def __init__(self, model, session=db_session):
        """

        :param session: sqlalchemy session instance
        :param model: sqlalchemy model
        """
        super().__init__(model)
        self._session = session
        self.support = self.support_class(self._model, self._session)

    def validate(self, schema):
        """

        :param schema: schema compatible with self validator
        :return:
        """
        schema = schema() if callable(schema) else schema
        return self.validator.validate(schema)

    def create_resource(self, data):
        """

        :param data: dictionary data that represents the resource
        :return: sqlalchemy model instance
        """
        return self._model(**data)

    # noinspection PyMethodMayBeStatic
    def update_resource(self, resource, data):
        """

        :param resource: sqlalchemy model instance
        :param data: dictionary data that represents the resource
        :return:
        """
        resource.update(data)
        return resource

    def _session_exception_handler(self, exception: SQLAlchemyError) -> t.NoReturn:
        cap.logger.exception(exception)
        self._session.rollback()

        if isinstance(exception, DBError):
            abort(httpcode.CONFLICT, response={"cause": exception.as_dict()})
        abort(httpcode.INTERNAL_SERVER_ERROR)

    def _create(self, res):
        try:
            self._session.add(res)
            self._session.commit()
        except SQLAlchemyError as exc:
            self._session_exception_handler(exc)

    def _update(self, res):
        try:
            self._session.merge(res)
            self._session.commit()
        except SQLAlchemyError as exc:
            self._session_exception_handler(exc)

    @classmethod
    def _prepare_upsert_filters(cls, *_, **__) -> t.Dict[str, t.Any]:
        return {}

    # pylint: disable=inconsistent-return-statements
    def _upsert(self, data) -> t.Tuple[t.Any, int]:
        try:
            res, created = self.support.update_or_create(
                data, **self._prepare_upsert_filters(data)
            )
            self._session.commit()
            return res, httpcode.CREATED if created else httpcode.SUCCESS
        except SQLAlchemyError as exc:
            self._session_exception_handler(exc)

    def on_post(self, *_, **__) -> t.Tuple[t.Dict[str, t.Any], int]:
        payload = self.validate(self.post_schema)
        res = self.create_resource(payload)
        self._create(res)
        return res.to_dict(), httpcode.CREATED

    def _delete(self, res_id, *_, **__) -> t.Dict[str, t.Any]:
        """

        :param res_id: resource identifier (primary key value)
        """
        res = self._model.query.get_or_404(res_id)
        data = res.to_dict()

        try:
            self._session.delete(res)
            self._session.commit()
        except SQLAlchemyError as exc:
            self._session_exception_handler(exc)

        return data

    def on_delete(self, res_id, *args, **kwargs):
        self._delete(res_id, *args, **kwargs)

    def on_put(self, *_, res_id=None, **__) -> t.Tuple[t.Dict[str, t.Any], int]:
        """

        :param res_id: resource identifier (primary key value)
            it is optional in order to implement upsert
        :return:
        """
        payload = self.validate(self.put_schema or self.post_schema)

        if res_id is not None:
            res = self._model.query.get_or_404(res_id)
            res = self.update_resource(res, payload)
            self._update(res)
            return res.to_dict(), httpcode.SUCCESS

        res, status = self._upsert(payload)
        return res.to_dict(), status


class PatchApiView(Restful):
    methods_subresource = None
    methods_collection = None
    methods_resource = None
    methods: t.ClassVar[t.Optional[t.Collection[str]]] = [
        HttpMethod.PATCH,
    ]

    @classmethod
    def register(
        cls,
        app,
        name: t.Optional[str] = None,
        urls: UrlsType = (),
        view: t.Optional[t.Type[BaseView]] = None,
        **kwargs,
    ) -> t.Callable:
        _view = t.cast(t.Type[BaseView], view or cls)
        return BaseView.register(app, name, urls, _view, **kwargs)

    @builder.no_content
    def patch(self, *args, **kwargs):
        return self.on_patch(*args, **kwargs)

    def on_patch(self, *_, **__):
        return self.not_implemented()



import typing as t

import argon2
from vbcore.crypto.argon import Argon2 as HasherArgon2, Argon2Options


class Argon2:
    CONFIG_PREFIX = "ARGON2_"
    PROFILES: t.Dict[str, argon2.Parameters] = {
        "low": argon2.profiles.RFC_9106_LOW_MEMORY,
        "high": argon2.profiles.RFC_9106_HIGH_MEMORY,
    }

    def __init__(self, app=None, **kwargs):
        self._argon = None

        if app is not None:
            self.init_app(app, **kwargs)

    @classmethod
    def set_default_config(cls, app):
        app.config.setdefault(f"{cls.CONFIG_PREFIX}PROFILE", "low")
        app.config.setdefault(f"{cls.CONFIG_PREFIX}ENCODING", "utf-8")

        argon_profile = app.config.pop(f"{cls.CONFIG_PREFIX}PROFILE", None)
        if argon_profile not in cls.PROFILES:
            profiles = tuple(cls.PROFILES.values())
            raise ValueError(f"profile not allowed, choose one of: {profiles}")

        params = cls.PROFILES[argon_profile]
        app.config.setdefault(f"{cls.CONFIG_PREFIX}TIME_COST", params.time_cost)
        app.config.setdefault(f"{cls.CONFIG_PREFIX}HASH_LEN", params.hash_len)
        app.config.setdefault(f"{cls.CONFIG_PREFIX}MEMORY_COST", params.memory_cost)
        app.config.setdefault(f"{cls.CONFIG_PREFIX}PARALLELISM", params.parallelism)
        app.config.setdefault(f"{cls.CONFIG_PREFIX}SALT_LEN", params.salt_len)

    def init_app(self, app, argon_class: t.Type[HasherArgon2] = HasherArgon2):
        self.set_default_config(app)
        self._argon = argon_class(
            Argon2Options(**app.config.get_namespace(self.CONFIG_PREFIX))
        )
        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["argon2"] = self

    def __getattr__(self, item):
        return getattr(self._argon, item)



import typing as t
from decimal import Decimal

from werkzeug.routing import BaseConverter, FloatConverter, Map, UnicodeConverter


class ListConverter(BaseConverter):
    def __init__(self, map_name: Map, sep: str = "+"):
        """

        :param map_name:
        :param sep:
        """
        super().__init__(map_name)
        self._sep = sep

    def to_python(self, value: str) -> t.List[str]:
        """

        :param value:
        :return:
        """
        return value.split(self._sep)

    def to_url(self, value: t.List[str]) -> str:
        """

        :param value:
        :return:
        """
        return super().to_url(self._sep.join(value))


class DecimalConverter(FloatConverter):
    num_convert = Decimal  # type: ignore

    def __init__(
        self,
        map_name: "Map",
        min: t.Optional[float] = None,  # pylint: disable=redefined-builtin
        max: t.Optional[float] = None,  # pylint: disable=redefined-builtin
        signed: bool = True,
    ) -> None:
        super().__init__(map_name, min=min, max=max, signed=signed)


CONVERTERS = {
    "list": ListConverter,
    "str": UnicodeConverter,
    "decimal": DecimalConverter,
}



import typing as t
from functools import partial

import flask
from vbcore.db.support import SQLASupport
from vbcore.http import httpcode, HttpMethod
from vbcore.tester.asserter import Asserter
from vbcore.tester.helpers import build_url

from flaskel import TestClient
from flaskel.utils.datastruct import ConfigProxy

config = ConfigProxy()
schemas = ConfigProxy("SCHEMAS")
url_for = partial(flask.url_for, _external=True)


def load_sample_data(filename: str):
    SQLASupport.exec_from_file(
        config.SQLALCHEMY_DATABASE_URI, filename, echo=config.SQLALCHEMY_ECHO
    )


class ApiTester:
    def __init__(self, client: TestClient, mimetype: t.Optional[str] = None):
        self.client = client
        self.mimetype = mimetype

    def token_header(
        self, call: bool = False, **kwargs
    ) -> t.Union[t.Callable, t.Dict[str, str]]:
        if call is True:
            return get_access_token(self.client, **kwargs)
        return partial(get_access_token, self.client, **kwargs)

    def perform(
        self,
        url: t.Optional[str] = None,
        method: str = HttpMethod.GET,
        view: t.Optional[str] = None,
        schema: t.Optional[str] = None,
        status: t.Optional[int] = None,
        params: t.Optional[dict] = None,
        mimetype: t.Optional[str] = None,
        **kwargs,
    ):
        url, _ = build_url(url_for(view) if view else url, **(params or {}))
        response = self.client.open(url, method=method, **kwargs)

        if status is not None:
            Asserter.assert_status_code(response, status)
        else:
            Asserter.assert_status_code(
                response,
                in_range=True,
                code=(httpcode.SUCCESS - 1, httpcode.MULTIPLE_CHOICES - 1),
            )

        if response.data and (mimetype or self.mimetype):
            Asserter.assert_equals(response.mimetype, mimetype or self.mimetype)

        if schema:
            Asserter.assert_json_schema(response.json, schema)

        return response

    def post(
        self,
        url: t.Optional[str] = None,
        view: t.Optional[str] = None,
        schema: t.Optional[str] = None,
        status: t.Optional[int] = None,
        params: t.Optional[dict] = None,
        mimetype: t.Optional[str] = None,
        **kwargs,
    ):
        return self.perform(
            method=HttpMethod.POST,
            url=url,
            view=view,
            schema=schema,
            status=status,
            params=params,
            mimetype=mimetype,
            **kwargs,
        )

    def put(
        self,
        url: t.Optional[str] = None,
        view: t.Optional[str] = None,
        schema: t.Optional[str] = None,
        status: t.Optional[int] = None,
        params: t.Optional[dict] = None,
        mimetype: t.Optional[str] = None,
        **kwargs,
    ):
        return self.perform(
            method=HttpMethod.PUT,
            url=url,
            view=view,
            schema=schema,
            status=status,
            params=params,
            mimetype=mimetype,
            **kwargs,
        )

    def get(
        self,
        url: t.Optional[str] = None,
        view: t.Optional[str] = None,
        schema: t.Optional[str] = None,
        status: t.Optional[int] = None,
        params: t.Optional[dict] = None,
        mimetype: t.Optional[str] = None,
        **kwargs,
    ):
        return self.perform(
            method=HttpMethod.GET,
            url=url,
            view=view,
            schema=schema,
            status=status,
            params=params,
            mimetype=mimetype,
            **kwargs,
        )

    def delete(
        self,
        url: t.Optional[str] = None,
        view: t.Optional[str] = None,
        status: t.Optional[int] = None,
        params: t.Optional[dict] = None,
        mimetype: t.Optional[str] = None,
        **kwargs,
    ):
        return self.perform(
            method=HttpMethod.DELETE,
            url=url,
            view=view,
            schema=None,
            status=status,
            params=params,
            mimetype=mimetype,
            **kwargs,
        )

    def patch(
        self,
        url: t.Optional[str] = None,
        view: t.Optional[str] = None,
        schema: t.Optional[str] = None,
        status: t.Optional[int] = None,
        params: t.Optional[dict] = None,
        mimetype: t.Optional[str] = None,
        **kwargs,
    ):
        return self.perform(
            method=HttpMethod.PATCH,
            url=url,
            view=view,
            schema=schema,
            status=status,
            params=params,
            mimetype=mimetype,
            **kwargs,
        )

    def restful(
        self,
        view: str,
        res_id=None,
        headers: t.Optional[dict] = None,
        body_create: t.Optional[dict] = None,
        body_update: t.Optional[dict] = None,
        schema_read=None,
        schema_collection=None,
        methods: t.Tuple[str, ...] = (
            HttpMethod.GET,
            HttpMethod.POST,
            HttpMethod.PUT,
            HttpMethod.DELETE,
        ),
        **params,
    ):
        url_collection, args = build_url(url_for(view), **params)

        if HttpMethod.POST in methods:
            res = self.post(
                url_collection,
                json=body_create,
                headers=headers,
                status=httpcode.CREATED,
                schema=schema_read,
            )
            res_id = res.json.id

        url_resource = f"{url_for(view, res_id=res_id)}?{args}"

        if HttpMethod.GET in methods:
            self.get(
                url_collection,
                headers=headers,
                schema=schema_collection,
            )
            self.get(
                url_resource,
                headers=headers,
                schema=schema_read,
            )

        if HttpMethod.PUT in methods:
            self.put(
                url_resource,
                headers=headers,
                json=body_update or body_create,
                schema=schema_read,
            )

        if HttpMethod.DELETE in methods:
            self.delete(
                url_resource,
                headers=headers,
                status=httpcode.NO_CONTENT,
            )
            self.get(
                url_resource,
                headers=headers,
                status=httpcode.NOT_FOUND,
                schema=schemas.API_PROBLEM,
            )


def get_access_token(
    client,
    is_query: bool = False,
    token: t.Optional[str] = None,
    token_type: str = "Bearer",
    access_view: str = "auth.token_access",
    credentials: t.Optional[t.Union[t.Dict[str, str], t.Tuple[str, str]]] = None,
) -> t.Dict[str, str]:
    conf = client.application.config

    if token is not None:
        return dict(Authorization=f"{token_type} {token}")

    if not credentials:
        credentials = dict(email=conf.ADMIN_EMAIL, password=conf.ADMIN_PASSWORD)
    elif isinstance(credentials, tuple):
        credentials = dict(email=credentials[0], password=credentials[1])

    tokens = client.post(url_for(access_view), json=credentials)
    if is_query is True:
        return {conf.JWT_QUERY_STRING_NAME: tokens.json.access_token}

    return dict(Authorization=f"{token_type} {tokens.json.access_token}")



from ext import EXTENSIONS
from views import BLUEPRINTS, VIEWS

from flaskel import AppBuilder, middlewares as middle, Server

APP_CONFIG = dict(
    static_folder=None,  # because static is inside web blueprint
    blueprints=BLUEPRINTS,
    extensions=EXTENSIONS,
    views=VIEWS,
    middlewares=(
        middle.RequestID,
        middle.HTTPMethodOverride,
        middle.ReverseProxied,
    ),
)

factory = AppBuilder(**APP_CONFIG)  # type: ignore


def create_app(conf=None):
    """
    useful for external wsgi server

    :param conf:
    :return:
    """
    return factory.create(conf)


def cli():
    Server(factory).run_from_cli()


if __name__ == "__main__":
    cli()



import os

import decouple
from vbcore import yaml
from vbcore.configurator import config
from vbcore.date_helper import Seconds
from vbcore.http.headers import ContentTypeEnum, HeaderEnum

from flaskel.utils import logger

DEBUG = config("DEBUG", default=False, cast=bool)
FLASK_DEBUG = config("FLASK_DEBUG", default=DEBUG, cast=bool)
TESTING = config("TESTING", default=False, cast=bool)
APP_NAME = config("APP_NAME", default="flaskel")
APP_HOST = config("APP_HOST", default="127.0.0.1")
APP_PORT = config("APP_PORT", default=5000, cast=int)
FLASK_APP = config("FLASK_APP", default="app:app")
SERVER_NAME = config("SERVER_NAME", default=f"{APP_HOST}:{APP_PORT}")

FLASK_ENV = config(
    "FLASK_ENV",
    default="development" if DEBUG else "production",
    cast=decouple.Choices(["development", "production"]),
)

LOCALE = config("LOCALE", default="en_EN.utf8")
TEMPLATES_AUTO_RELOAD = config("TEMPLATES_AUTO_RELOAD", default=DEBUG, cast=bool)
EXPLAIN_TEMPLATE_LOADING = config("EXPLAIN_TEMPLATE_LOADING", default=False, cast=bool)
APIDOCS_ENABLED = config("APIDOCS_ENABLED", default=True, cast=bool)
CONF_PATH = config("CONF_PATH", default=os.path.join(config.search_path, "resources"))

JWT_DEFAULT_SCOPE = None
JWT_IDENTITY_CLAIM = "identity"
JWT_DEFAULT_TOKEN_TYPE = "bearer"
JWT_ERROR_MESSAGE_KEY = "message"
JWT_QUERY_STRING_NAME = "token"
JWT_TOKEN_LOCATION = ["headers", "query_string"]
JWT_ALGORITHM = config("JWT_ALGORITHM", default="HS512")
JWT_DECODE_ALGORITHMS = config(
    "JWT_ALGORITHM", default=JWT_ALGORITHM, cast=decouple.Csv()
)
JWT_ACCESS_TOKEN_EXPIRES = config(
    "JWT_ACCESS_TOKEN_EXPIRES", default=Seconds.day, cast=int
)
JWT_REFRESH_TOKEN_EXPIRES = config(
    "JWT_REFRESH_TOKEN_EXPIRES", default=Seconds.day * 14, cast=int
)

SQLALCHEMY_DATABASE_URI = config("DATABASE_URL", default="sqlite:///db.sqlite")

MONGO_URI = config("MONGO_URI", default="mongodb://localhost:27017")
MONGO_OPTS = {
    "connectTimeoutMS": config("MONGO_CONN_TIMEOUT_MS", default=100, cast=int),
    "serverSelectionTimeoutMS": config(
        "MONGO_SERVER_SELECTION_TIMEOUT_MS", default=100, cast=int
    ),
}

REDIS_URL = config("REDIS_URL", default="redis://127.0.0.1:6379")
REDIS_OPTS = {
    "decode_responses": True,
    "socket_connect_timeout": config("REDIS_CONN_TIMEOUT", default=0.05, cast=float),
}

_CORS_EXPOSE_HEADERS_DEFAULT = [
    HeaderEnum.X_PAGINATION_COUNT,
    HeaderEnum.X_PAGINATION_NUM_PAGES,
    HeaderEnum.X_PAGINATION_PAGE,
    HeaderEnum.X_PAGINATION_PAGE_SIZE,
    HeaderEnum.X_REQUEST_ID,
    HeaderEnum.X_API_VERSION,
    HeaderEnum.X_RATELIMIT_RESET,
    HeaderEnum.X_RATELIMIT_REMAINING,
    HeaderEnum.X_RATELIMIT_LIMIT,
    HeaderEnum.RETRY_AFTER,
]
CORS_EXPOSE_HEADERS = config(
    "CORS_EXPOSE_HEADERS",
    default=",".join(_CORS_EXPOSE_HEADERS_DEFAULT),
    cast=decouple.Csv(),
)

BASIC_AUTH_USERNAME = config("BASIC_AUTH_USERNAME", default="admin")
BASIC_AUTH_PASSWORD = config("BASIC_AUTH_PASSWORD", default="admin")

ADMIN_EMAIL = config("ADMIN_EMAIL", default="admin@mail.com")
ADMIN_PASSWORD = config("ADMIN_PASSWORD", default="admin")

MAIL_DEBUG = config("MAIL_DEBUG", default=DEBUG, cast=bool)
MAIL_SERVER = config("MAIL_SERVER", default="127.0.0.1")
MAIL_PORT = config("MAIL_PORT", default=25, cast=int)
MAIL_USERNAME = config("MAIL_USERNAME", default="")
MAIL_PASSWORD = config("MAIL_PASSWORD", default="")
MAIL_USE_SSL = config("MAIL_USE_SSL", default=False, cast=bool)
MAIL_USE_TLS = config("MAIL_USE_TLS", default=False, cast=bool)
MAIL_DEFAULT_SENDER = config("MAIL_DEFAULT_SENDER", default=ADMIN_EMAIL)
MAIL_DEFAULT_RECEIVER = config("MAIL_DEFAULT_RECEIVER", default=ADMIN_EMAIL)
MAIL_RECIPIENT = config("MAIL_RECIPIENT", default=ADMIN_EMAIL)
MAIL_TIMEOUT = config("MAIL_TIMEOUT", default=60, cast=int)

PREFERRED_URL_SCHEME = config(
    "PREFERRED_URL_SCHEME", default="http" if FLASK_ENV == "development" else "https"
)

LOG_BUILDER = config("LOG_BUILDER", default="text")
LOG_APP_NAME = config("LOG_APP_NAME", default=APP_NAME)
LOG_LOGGER_NAME = config("LOG_LOGGER_NAME", default=FLASK_ENV)
LOG_REQ_HEADERS = config("LOG_REQ_HEADERS", default="", cast=decouple.Csv())
LOG_RESP_HEADERS = config("LOG_RESP_HEADERS", default="", cast=decouple.Csv())
LOG_REQ_SKIP_DUMP = config("LOG_REQ_SKIP_DUMP", default=not TESTING, cast=bool)
LOG_RESP_SKIP_DUMP = config("LOG_RESP_SKIP_DUMP", default=not TESTING, cast=bool)
LOG_REQ_FORMAT = config(
    "LOG_REQ_FORMAT",
    default="INCOMING REQUEST {address} {method} {scheme} {path}{headers}{body}",
)
LOG_RESP_FORMAT = config(
    "LOG_RESP_FORMAT",
    default="OUTGOING RESPONSE for {address} at {path}: STATUS {status}{headers}{body}",
)

CF_STRICT_ACCESS = config("CF_STRICT_ACCESS", default=False, cast=bool)
VERSION_STORE_MAX = config("VERSION_STORE_MAX", default=6, cast=int)
VERSION_CACHE_EXPIRE = config("VERSION_CACHE_EXPIRE", default=Seconds.hour, cast=int)

HTTP_TIMEOUT = config("HTTP_TIMEOUT", default=10, cast=int)
HTTP_SSL_VERIFY = config("HTTP_SSL_VERIFY", default=True, cast=bool)
HTTP_PROTECT_BODY = config("HTTP_PROTECT_BODY", default=False, cast=bool)
HTTP_DUMP_BODY = [
    config("HTTP_DUMP_REQ_BODY", default=False, cast=bool),
    config("HTTP_DUMP_RESP_BODY", default=False, cast=bool),
]

MAX_CONTENT_LENGTH = config("MAX_CONTENT_LENGTH", default=10 * 10**6, cast=int)
USE_X_SENDFILE = config("USE_X_SENDFILE", default=not DEBUG, cast=bool)
ENABLE_ACCEL = config("ENABLE_ACCEL", default=True, cast=bool)
ACCEL_BUFFERING = True
ACCEL_CHARSET = "utf-8"
ACCEL_LIMIT_RATE = "off"

WSGI_WERKZEUG_PROFILER_FILE = config(
    "WSGI_WERKZEUG_PROFILER_FILE", default="profiler.txt"
)
WSGI_WERKZEUG_PROFILER_RESTRICTION = config(
    "WSGI_WERKZEUG_PROFILER_RESTRICTION",
    default="0.1",
    cast=decouple.Csv(post_process=tuple),
)
WSGI_WERKZEUG_LINT_ENABLED = config(
    "WSGI_WERKZEUG_LINT_ENABLED", default=TESTING, cast=bool
)
WSGI_WERKZEUG_PROFILER_ENABLED = config(
    "WSGI_WERKZEUG_PROFILER_ENABLED", default=TESTING, cast=bool
)

SQLALCHEMY_TRACK_MODIFICATIONS = False
SQLALCHEMY_ECHO = config("SQLALCHEMY_ECHO", default=TESTING, cast=bool)

REQUEST_ID_HEADER = HeaderEnum.X_REQUEST_ID
RATELIMIT_ENABLED = config("RATELIMIT_ENABLED", default=not DEBUG, cast=bool)
RATELIMIT_HEADERS_ENABLED = config("RATELIMIT_HEADERS_ENABLED", default=True, cast=bool)
RATELIMIT_IN_MEMORY_FALLBACK_ENABLED = config(
    "RATELIMIT_IN_MEMORY_FALLBACK_ENABLED", default=True, cast=bool
)

DATABASE_URL = SQLALCHEMY_DATABASE_URI
ERROR_PAGE = "core/error.html"
SECRET_KEY_FILE_NAME = ".secret.key"
SECRET_KEY_MIN_LENGTH = 256
RB_DEFAULT_ACCEPTABLE_MIMETYPES = [
    ContentTypeEnum.JSON,
]

PRETTY_DATE = "%d %B %Y %I:%M %p"
DATE_ISO_FORMAT = "%Y-%m-%dT%H:%M:%S"
SEND_FILE_MAX_AGE_DEFAULT = config(
    "SEND_FILE_MAX_AGE_DEFAULT", default=Seconds.day, cast=int
)

JSONRPC_BATCH_MAX_REQUEST = config("JSONRPC_BATCH_MAX_REQUEST", default=10, cast=int)

RATELIMIT_STORAGE_URL = config("RATELIMIT_STORAGE_URL", default=REDIS_URL)
RATELIMIT_KEY_PREFIX = config("RATELIMIT_KEY_PREFIX", default=APP_NAME)
RATELIMIT_STORAGE_OPTIONS = {
    "socket_timeout": REDIS_OPTS["socket_connect_timeout"],
    "socket_connect_timeout": REDIS_OPTS["socket_connect_timeout"],
}

IPBAN_ENABLED = config("IPBAN_ENABLED", default=True, cast=bool)
IPBAN_COUNT = config("IPBAN_COUNT", default=20, cast=int)
IPBAN_SECONDS = config("IPBAN_SECONDS", default=Seconds.hour, cast=int)
IPBAN_STATUS_CODE = config("IPBAN_STATUS_CODE", default=403, cast=int)
IPBAN_CHECK_CODES = config(
    "IPBAN_CHECK_CODES",
    default="404,405,501",
    cast=decouple.Csv(post_process=lambda items: tuple(int(i) for i in items)),
)

LIMITER = {
    "FAIL": "1/second",
    "FAST": "30/minute",
    "MEDIUM": "20/minute",
    "SLOW": "10/minute",
    "BYPASS_KEY": "X-Limiter-Bypass",
    "BYPASS_VALUE": "bypass-rate-limit",
}

SCHEDULER_JOBSTORES = {"default": {"type": "memory"}}
SCHEDULER_JOB_DEFAULTS = {"coalesce": False, "max_instances": 10}
SCHEDULER_EXECUTORS = {"default": {"type": "threadpool", "max_workers": 20}}
SCHEDULER_AUTO_START = config("SCHEDULER_AUTO_START", default=True, cast=bool)
SCHEDULER_API_ENABLED = config("SCHEDULER_API_ENABLED", default=False, cast=bool)

CACHE_TYPE = "flask_caching.backends.redis"
CACHE_REDIS_URL = config("CACHE_REDIS_URL", default=REDIS_URL)
CACHE_DISABLED = config("CACHE_DISABLED", default=False, cast=bool)
CACHE_DEFAULT_TIMEOUT = config("CACHE_DEFAULT_TIMEOUT", default=Seconds.hour, cast=int)
CACHE_KEY_PREFIX = config("CACHE_KEY_PREFIX", default=APP_NAME)
CACHE_OPTIONS = {
    "socket_timeout": REDIS_OPTS["socket_connect_timeout"],
    "socket_connect_timeout": REDIS_OPTS["socket_connect_timeout"],
}

APISPEC = yaml.load_optional_yaml_file(
    os.path.join(CONF_PATH, "swagger.yaml"), debug=DEBUG
)
SCHEMAS = yaml.load_optional_yaml_file(
    os.path.join(CONF_PATH, "schemas.yaml"), debug=DEBUG
)
SCHEDULER_JOBS = yaml.load_optional_yaml_file(
    os.path.join(CONF_PATH, "scheduler.yaml"), debug=DEBUG
)
LOGGING = yaml.load_optional_yaml_file(
    os.path.join(CONF_PATH, "log.yaml"), default=logger.LOGGING
)



import functools
import typing as t
from base64 import b64encode

from flask_caching import Cache as FlaskCache
from vbcore.http import httpcode, HttpMethod
from vbcore.http.headers import HeaderEnum

from flaskel.flaskel import cap, request, Response


class Caching(FlaskCache):
    def __init__(
        self,
        app=None,
        with_jinja2_ext: bool = True,
        config: t.Optional[dict] = None,
        key_separator: t.Union[t.Callable, str] = "/",
        key_prefix: t.Union[t.Callable, str] = "/cached_request",
        default_timeout: t.Union[t.Callable, str] = None,
        cache_control_bypass: t.Union[t.Callable, str] = "no-store",
        headers_in_keys: t.Union[t.Callable, t.Iterable[str]] = (
            HeaderEnum.CONTENT_TYPE,
        ),
        cacheable_methods: t.Union[t.Callable, t.Iterable[str]] = (HttpMethod.GET,),
    ):
        self.key_prefix = key_prefix
        self.key_separator = key_separator
        self.default_timeout = default_timeout
        self.headers_in_keys = headers_in_keys
        self.cacheable_methods = cacheable_methods
        self.cache_control_bypass = cache_control_bypass
        super().__init__(app, with_jinja2_ext, config)

    @staticmethod
    def optional_callable(key):
        return key() if callable(key) else key

    @staticmethod
    def hash_method(data: str) -> str:
        encoding = "utf-8"
        return b64encode(data.encode(encoding)).decode(encoding)

    @staticmethod
    def make_query_string() -> str:
        tokens: t.List[str] = []
        query = (pair for pair in request.args.items(multi=True))
        for param, value in query:
            tokens.extend(f"{param}={v}" for v in value)
        return "&".join(sorted(tokens))

    @staticmethod
    def make_headers_string(headers: t.List[str], separator: str) -> str:
        tokens: t.List[str] = []
        for header in headers:
            value = request.headers.get_all(header)
            tokens.extend(value[0].split(", ") if len(value) == 1 else value)
        return separator.join((sorted(tokens)))

    def unless(self, *_, **__) -> bool:
        if cap.config.CACHE_DISABLED:
            return True

        bypass = self.optional_callable(self.cache_control_bypass)
        return bypass == request.headers.get(HeaderEnum.CACHE_CONTROL)

    def response_filter(self, response) -> bool:
        if request.method not in self.optional_callable(self.cacheable_methods):
            return False
        if isinstance(response, Response):
            return httpcode.is_ok(response.status_code)

        cap.logger.debug(
            "response is not of type Response, so no status code can be checked"
        )
        return False

    def make_cache_key(self, *_, **__) -> str:
        """
        Returns a string formatted as follow:
            <prefix><sep><url><sep>[<query><sep><headers>]

        string between [] is hashed with ``hash_method``
        note that query string and headers are sorted
        """
        key_prefix = self.optional_callable(self.key_prefix)
        separator = self.optional_callable(self.key_separator)
        url = request.base_url.rstrip(separator)
        query = self.make_query_string()
        headers = self.optional_callable(self.headers_in_keys)
        headers = self.make_headers_string(headers, separator)
        hashed_key = self.hash_method(f"{query}{separator}{headers}")
        return f"{key_prefix}{separator}{url}{separator}{hashed_key}"

    def superclass_cached(self, **kwargs):
        """This is here only to mock it in tests, until a better way is found"""
        return super().cached(**kwargs)

    def cached(self, **kwargs):
        def wrapper(f):
            @functools.wraps(f)
            def decorator(*args, **kw):
                kwargs.setdefault("unless", self.unless)
                kwargs.setdefault("timeout", self.default_timeout)
                kwargs.setdefault("make_cache_key", self.make_cache_key)
                kwargs.setdefault("response_filter", self.response_filter)
                kwargs["timeout"] = self.optional_callable(kwargs["timeout"])
                return self.superclass_cached(**kwargs)(f)(*args, **kw)

            return decorator

        return wrapper



import dataclasses
import functools
import typing as t
from datetime import timedelta

import sqlalchemy as sa
from flask_httpauth import HTTPBasicAuth
from flask_jwt_extended import (
    create_access_token,
    create_refresh_token,
    decode_token,
    get_jwt,
    get_jwt_identity,
    jwt_required,
    JWTManager,
)
from vbcore.base import BaseDTO
from vbcore.datastruct import ObjectDict
from vbcore.db.mixins import StandardMixin
from vbcore.http import httpcode

from flaskel.flaskel import cap
from flaskel.http.exceptions import abort

token_auth: JWTManager = JWTManager()
basic_auth: HTTPBasicAuth = HTTPBasicAuth()


@basic_auth.verify_password
def simple_basic_auth(username: str, password: str) -> t.Optional[dict]:
    if (
        username == cap.config.BASIC_AUTH_USERNAME
        and password == cap.config.BASIC_AUTH_PASSWORD
    ):
        return dict(username=username, password=password)
    return None


@token_auth.invalid_token_loader
def invalid_token_loader(mess) -> t.Tuple[dict, int]:
    return dict(message=mess), httpcode.UNAUTHORIZED


class RevokedTokenMixin(StandardMixin):
    jti = sa.Column(sa.String(120), nullable=False, unique=True)

    def __repr__(self):
        return f"<RevokedToken: {self.id} - {self.jti}>"

    @classmethod
    def is_block_listed(cls, jti: str) -> bool:
        # noinspection PyUnresolvedReferences
        return bool(cls.query.filter_by(jti=jti).first())


@dataclasses.dataclass(frozen=True)
class TokenInfo(BaseDTO):
    fresh: bool
    expires_in: int
    issued_at: int
    token_type: str
    type: str
    jti: str
    identity: dict
    scope: t.Optional[str] = None


@dataclasses.dataclass(frozen=True)
class TokenData(BaseDTO):
    access_token: str
    expires_in: int
    issued_at: int
    token_type: str
    scope: t.Optional[str] = None
    refresh_token: t.Optional[str] = None

    def to_dict(self, **kwargs) -> ObjectDict:
        data = ObjectDict(
            access_token=self.access_token,
            expires_in=self.expires_in,
            issued_at=self.issued_at,
            token_type=self.token_type,
            scope=self.scope,
            **kwargs,
        )
        if self.refresh_token:
            data.refresh_token = self.refresh_token
        return data


class BaseTokenHandler:
    def __init__(self, blocklist_loader: t.Optional[t.Callable] = None):
        token_auth.token_in_blocklist_loader(blocklist_loader)

    def check_token_block_listed(self, jwt_headers, jwt_data) -> bool:
        _ = jwt_headers, jwt_data
        return False

    def revoke(self, token: t.Optional[str] = None):
        """can be implemented by subclass"""

    @classmethod
    def role(cls):
        return None

    # noinspection PyMethodMayBeStatic
    def prepare_identity(self, data):
        return data

    @classmethod
    def check_permission(cls, role):
        if cls.role() != role:
            abort(httpcode.FORBIDDEN)

    @classmethod
    def auth_required(cls, perm=None, **kwargs):
        def wrapper(fun):
            @functools.wraps(fun)
            @jwt_required(**kwargs)
            def check_optional_permission(*args, **kw):
                if perm is not None:
                    cls.check_permission(perm)
                return fun(*args, **kw)

            return check_optional_permission

        return wrapper

    @classmethod
    def identity(cls) -> ObjectDict:
        identity = get_jwt_identity()
        if isinstance(identity, dict):
            identity = ObjectDict(**identity)
        return identity

    @classmethod
    def get_raw(cls) -> ObjectDict:
        return ObjectDict(**get_jwt())

    @classmethod
    def decode(cls, token: str) -> ObjectDict:
        return ObjectDict(**decode_token(token))

    def get_access(self, identity=None, expires: t.Optional[int] = None) -> str:
        return create_access_token(
            identity=identity or self.identity(),
            expires_delta=timedelta(seconds=expires) if expires else None,
        )

    def get_refresh(self, identity=None, expires: t.Optional[int] = None) -> str:
        return create_refresh_token(
            identity=identity or self.identity(),
            expires_delta=timedelta(seconds=expires) if expires else None,
        )

    def refresh(self, expires: t.Optional[int] = None) -> TokenData:
        access_token = self.get_access(expires=expires)
        decoded = self.decode(access_token)
        return TokenData(
            access_token=access_token,
            expires_in=decoded.exp,
            issued_at=decoded.iat,
            token_type=cap.config.JWT_DEFAULT_TOKEN_TYPE,
            scope=cap.config.JWT_DEFAULT_SCOPE,
        )

    def create(
        self,
        identity,
        refresh: bool = True,
        expires_access: t.Optional[int] = None,
        expires_refresh: t.Optional[int] = None,
        scope: t.Optional[str] = None,
    ) -> TokenData:
        identity = self.prepare_identity(identity)
        access_token = self.get_access(identity=identity, expires=expires_access)
        decoded = self.decode(access_token)
        return TokenData(
            access_token=access_token,
            expires_in=decoded.exp,
            issued_at=decoded.iat,
            token_type=cap.config.JWT_DEFAULT_TOKEN_TYPE,
            scope=scope or cap.config.JWT_DEFAULT_SCOPE,
            refresh_token=self.get_refresh(identity=identity, expires=expires_refresh)
            if refresh
            else None,
        )

    def dump(self) -> TokenInfo:
        data = self.get_raw()
        return TokenInfo(
            fresh=data.fresh,
            type=data.type,
            token_type=data.token_type,
            scope=data.scope,
            expires_in=data.exp,
            issued_at=data.iat,
            jti=data.jti,
            identity=data.identity,
        )


class DBTokenHandler(BaseTokenHandler):
    def __init__(self, model, session, blocklist_loader: t.Optional[t.Callable] = None):
        self.model = model
        self.session = session
        super().__init__(blocklist_loader or self.check_token_block_listed)

    def check_token_block_listed(self, jwt_headers, jwt_data) -> bool:
        return self.model.is_block_listed(jwt_data["jti"])

    def revoke(self, token: t.Optional[str] = None):
        decoded_token = self.decode(token) if token else self.get_raw()
        if decoded_token.jti:
            self.session.add(self.model(jti=decoded_token.jti))
            self.session.commit()


class RedisTokenHandler(BaseTokenHandler):
    entry_value = "true"
    key_prefix = "token_revoked::"

    def __init__(self, redis, blocklist_loader: t.Optional[t.Callable] = None):
        self.redis = redis
        super().__init__(blocklist_loader or self.check_token_block_listed)

    def check_token_block_listed(self, jwt_headers, jwt_data) -> bool:
        entry = self.redis.get(f"{self.key_prefix}{jwt_data['jti']}")
        if not entry:
            return False

        if isinstance(entry, str):
            return entry == self.entry_value
        return entry == self.entry_value.encode()

    def revoke(self, token: t.Optional[str] = None):
        decoded_token = self.decode(token) if token else self.get_raw()
        self.redis.set(f"{self.key_prefix}{decoded_token.jti}", self.entry_value)



from functools import partial

from vbcore.http import httpcode
from webargs import fields, flaskparser

from flaskel.http.exceptions import abort

parser = flaskparser.FlaskParser()
query = partial(parser.use_args, location="query")
payload = partial(parser.use_args, location="json")


class Field:
    integer = partial(fields.Integer)
    string = partial(fields.String)
    decimal = partial(fields.Decimal)
    boolean = partial(fields.Boolean, truthy={"true", 1}, falsy={"false", 0})
    positive = partial(integer, validate=lambda x: x > 0)
    negative = partial(integer, validate=lambda x: x < 0)
    not_positive = partial(integer, validate=lambda x: x <= 0)
    not_negative = partial(integer, validate=lambda x: x >= 0)
    isodate = partial(fields.DateTime)
    str_list = partial(fields.DelimitedList, cls_or_instance=string(), delimiter=",")
    int_list = partial(fields.DelimitedList, cls_or_instance=integer(), delimiter=",")


class OptField:
    integer = partial(Field.integer, load_default=None)
    string = partial(Field.string, load_default=None)
    decimal = partial(Field.decimal, load_default=None)
    boolean = partial(Field.boolean, load_default=False)
    positive = partial(Field.positive, load_default=None)
    negative = partial(Field.negative, load_default=None)
    not_positive = partial(Field.not_positive, load_default=None)
    not_negative = partial(Field.not_negative, load_default=None)
    isodate = partial(Field.isodate, load_default=None)
    str_list = partial(Field.str_list, load_default=())
    int_list = partial(Field.int_list, load_default=())


class ReqField:
    integer = partial(Field.integer, required=True)
    string = partial(Field.string, required=True)
    decimal = partial(Field.decimal, required=True)
    boolean = partial(Field.boolean, required=True)
    positive = partial(Field.positive, required=True)
    negative = partial(Field.negative, required=True)
    not_positive = partial(Field.not_positive, required=True)
    not_negative = partial(Field.not_negative, required=True)
    isodate = partial(Field.isodate, required=True)
    str_list = partial(Field.str_list, required=True)
    int_list = partial(Field.int_list, required=True)


@parser.error_handler
def handle_error(error, *_, **__):
    abort(httpcode.BAD_REQUEST, response=error.messages)


query_paginate = partial(
    query,
    dict(
        page=OptField.positive(),
        page_size=OptField.positive(),
        related=OptField.boolean(load_default=False),
    ),
)


def paginate(f=None):
    @query_paginate()
    def _route(*args, **kwargs):
        params = args[-1]
        if f is None:
            return params

        return f(*args[:-1], params=params, **kwargs)

    return _route if f else _route()



import typing as t

from gevent.pywsgi import WSGIServer  # pylint: disable=import-error

from flaskel import Flaskel

from .base import BaseApplication


class WSGIGevent(BaseApplication, WSGIServer):
    def __init__(self, app: Flaskel, options: t.Optional[dict] = None):
        BaseApplication.__init__(self, app, options)
        WSGIServer.__init__(
            self,
            self._bind,
            self.application,
            spawn=options.get("spawn") or "default",
            backlog=options.get("backlog"),
        )

    def run(self):
        self.serve_forever()
